"""Operators"""

from __future__ import annotations

# from itertools import islice
from typing import TYPE_CHECKING

from ..sets.cases import Elem, FCase
from ..sets.function import F

if TYPE_CHECKING:
    from ..sets.index import I
    from ..sets.variable import V


def sigma(variable: V, over: I = None, position: int = None) -> F:
    """Summation, allows better printing, avoids recurssion error

    Args:
        variable (V): variable set
        over (I): over what index
        position (int, optional): position of the index in the variable indices. Defaults to None.
        neg (bool, optional): if this is a negative sum. Defaults to False.
    Returns:
        F: summed up function
    """

    if over:
        length = len(over)

        if not position:
            position = variable.index.index(over)

        # Precompute slices
        before = variable.index[:position]
        after = variable.index[position + 1 :]

        # Build variables
        _variables = [
            variable(*before, _index, *after, make_new=True) for _index in over
        ]

    else:
        # sum over the entire set
        _variables = variable._
        length = len(variable)
        position = None
        over = variable.index

    if length == 2:
        # this checks for v_0 + v_1
        return _variables[0] + _variables[1]

    f = F()

    f.case = FCase.SUM

    f.issumhow = (variable.copy(), over, position)

    f.variables = _variables
    f.index = tuple(v.index for v in f.variables)
    f.one = f
    f.one_type = Elem.F
    f.give_name()
    # f.two_type = Elem.V
    f.rhs_thetas = []
    length_var = len(_variables[0])

    keys = list(zip(*(v.map for v in f.variables)))

    f.A = [[1] * length for _ in range(length_var)]

    for n in range(length_var):
        # make the child functions
        f_child = F()

        f_child.variables = [v[n] for v in f.variables]
        f_child.P = [v.n for v in f_child.variables]
        f_child.A = [1] * length

        key = keys[n]

        f_child.issumhow = (variable[length * n], over, position)
        f_child.parent = f
        f_child.case = FCase.SUM
        f_child.rhs_thetas = []
        f.P.append(f_child.P)
        f_child.give_name()
        f_child._ = [f_child]
        f._.append(f_child)
        f.map[key] = f_child
        f_child.map[key] = f_child
        f_child.parent = f
        f_child.index = key
        f_child.one = f_child
        f_child.one_type = Elem.F

    return f



"""Special Cases"""

from enum import Enum


class Elem(Enum):
    """Type of function"""

    # if element is a variable
    V = 'variable'
    # if element is a parameter
    P = 'parameter'
    # if element is a theta
    T = 'theta'
    # if element is a function
    F = 'function'


class FCase(Enum):
    """function cases"""

    # Negative variable (-1*v)
    NEGVAR = 'negative_variable'
    # sum of variables over one index element v0 + v1 + v2 + ... + vn
    SUM = 'summation'
    # Negative sum, - (v0 + v1 + v2 + ... + vn)
    NEGSUM = 'negative_summation'

    # Variable, never used to describe a function
    # rather a variable, is always a special case
    # of a function that is just a variable
    VAR = 'variable'

    # once changed, the case becomes
    FVAR = 'variable_as_function'

    # calculation
    # prints differently (f_cal = p*v)
    CALC = 'calculation'

    # when a variable is a function
    VARF = 'function_as_variable'


class PCase(Enum):
    """parameter cases"""

    # just a number, stretched into an array
    NUM = 'number'
    # just a negative number, stretched into an array
    NEGNUM = 'negative_number'
    # zero
    ZERO = 'zero'
    # set of numbers
    SET = 'set'
    # negated set
    NEGSET = 'negative_set'


class ICase(Enum):
    """index cases"""

    # DUMMY INDEX
    DUMMY = 'dummy_index'
    # SELF CONTAINED INDEX
    SELF = 'self_contained_index'



"""Solution"""

from dataclasses import dataclass

from ..sets.index import I
from ..sets.variable import V

try:
    from matplotlib import pyplot as plt
    from matplotlib import rc

    has_matplotlib = True

except ImportError:
    has_matplotlib = False


@dataclass
class Solution:
    """A State with its variables filled in"""

    name: str = " "

    def __post_init__(self):

        self._: dict[str, dict[str, list]] = {}

    def asdict(self):
        """Return the solution as a dictionary"""
        return {v: values['values'] for v, values in self._.items()}

    def update(self, variable_sets: list[V], n_sol: int = 0):
        """Add variables to the solution"""
        for v in variable_sets:
            if v.parent.name not in self._:
                self._[v.parent.name] = {
                    'latex': [],
                    'index_latex': [],
                    'positions': [],
                    'n': [],
                    'values': [],
                    'index': [],
                }
            self._[v.parent.name]['latex'].append(v.latex())
            self._[v.parent.name]['index_latex'].append(r'$' + v.latex(True) + r'$')
            self._[v.parent.name]['positions'].append(v.pos)
            self._[v.parent.name]['n'].append(v.n)
            self._[v.parent.name]['values'].append(v.X[n_sol])
            self._[v.parent.name]['index'].append(
                {
                    idx.name: {par.name: pos for par, pos in zip(idx.parent, idx.pos)}
                    for idx in v.index
                    if isinstance(idx, I)
                }
            )

    def __call__(self, variable: V):

        if variable.parent:
            return {
                'latex': self._[variable.parent.name]['latex'][variable.pos],
                'index_latex': r'$'
                + self._[variable.parent.name]['index_latex'][variable.pos]
                + r'$',
                'positions': self._[variable.parent.name]['positions'][variable.pos],
                'n': self._[variable.parent.name]['n'][variable.pos],
                'values': self._[variable.parent.name]['values'][variable.pos],
                'index': self._[variable.parent.name]['index'][variable.pos],
            }

        _return = {
            'latex': [],
            'index_latex': [],
            'positions': [],
            'n': [],
            'values': [],
            'index': [],
        }
        for var in variable:
            _return['latex'].append(self._[var.parent.name]['latex'][var.pos])
            _return['index_latex'].append(
                self._[var.parent.name]['index_latex'][var.pos]
            )
            _return['positions'].append(self._[var.parent.name]['positions'][var.pos])
            _return['n'].append(self._[var.parent.name]['n'][var.pos])
            _return['values'].append(self._[var.parent.name]['values'][var.pos])
            _return['index'].append(self._[var.parent.name]['index'][var.pos])
        return _return

    def draw(
        self,
        variable: V,
        kind: str = 'line',
        font_size: float = 16,
        fig_size: tuple[float, float] = (10, 6),
        linewidth: float = 0.7,
        color: str = 'blue',
        grid_alpha: float = 0.3,
        usetex: bool = True,
        x_ticks_lim: int = 20,
    ):
        """Plot the variable set

        Args:
            kind (str, optional): Type of plot ['line', 'bar']. Defaults to 'line'.
            font_size (float, optional): Font size for the plot. Defaults to 16.
            fig_size (tuple[float, float], optional): Size of the figure. Defaults to (12, 6).
            linewidth (float, optional): Width of the line in the plot. Defaults to 0.7.
            color (str, optional): Color of the line in the plot. Defaults to 'blue'.
            grid_alpha (float, optional): Transparency of the grid lines. Defaults to 0.3.
            usetex (bool, optional): Use LaTeX for text rendering. Defaults to True.


        """

        if not has_matplotlib:
            print(
                "matplotlib is an optional dependency, pip install gana[all] to get optional dependencies"
            )
            return

        data = self(variable)

        def check_whats_changing():
            check_unique: dict[int, list] = {}
            for index in data['index']:
                for n, idx in enumerate(index):
                    if n not in check_unique:
                        check_unique[n] = []
                    if idx not in check_unique[n]:
                        check_unique[n].append(idx)
            _hold = []
            _free = []

            for unq in check_unique.values():
                if len(unq) == 1:
                    _hold.append(unq[0])
                else:
                    _free.append(unq)

            return _hold, _free

        hold, free = check_whats_changing()

        ax = plt.subplots(figsize=fig_size)[1]

        # the indices are the x-axis
        x = range(len(data['values']))
        # the values are the y-axis
        y = data["values"]

        if usetex:
            rc(
                'font',
                **{'family': 'serif', 'serif': ['Computer Modern'], 'size': font_size},
            )
            rc('text', usetex=usetex)
        else:
            rc('font', **{'size': font_size})

        if kind == 'line':
            ax.plot(x, y, linewidth=linewidth, color=color)

        elif kind == 'bar':
            ax.bar(x, y, linewidth=linewidth, color=color)

        title = rf'${variable.tag or variable.name}$'

        if len(hold) > 0:
            title += r' ('
            for h in hold:
                title += rf'{h}, '

            title = title[:-2] + r')'

        ax.set_title(title)
        ax.set_ylabel(r'Values')

        ax.set_xlabel(r'Indices')
        ax.grid(alpha=grid_alpha)

        if len(x) <= x_ticks_lim:
            ax.set_xticks(x)

            _xtick_labels = []
            for idx_latex, idx in zip(data['index_latex'], data['index']):
                for i in idx:
                    if i in hold:
                        idx_latex = idx_latex.replace(i + ",", '').replace(i, '')
                _xtick_labels.append(idx_latex)

            ax.set_xticklabels(_xtick_labels)

        plt.rcdefaults()

    def draw_line(self, **kwargs):
        """Alias for plot with kind='line'"""
        self.draw(kind='line', **kwargs)

    def draw_bar(self, **kwargs):
        """Alias for plot with kind='bar'"""
        self.draw(kind='bar', **kwargs)



"""A set of index elements (X)"""

from operator import is_
from typing import Self

from IPython.display import Math, display

from .cases import ICase

try:
    from pyomo.environ import RangeSet as PyoRangeSet
    from pyomo.environ import Set as PyoSet

    has_pyomo = True
except ImportError:
    has_pyomo = False

try:
    from sympy import FiniteSet

    has_sympy = True
except ImportError:
    has_sympy = False


class I:
    """
    Set of index elements (X).

    :param members: Members of the Index set.
    :type members: str | int, optional
    :param size: Size of the Index set, creates an ordered set if given.
    :type size: int, optional
    :param mutable: If the Index set is mutable. Defaults to False.
    :type mutable: bool, optional
    :param tag: Tag/details. Defaults to None.
    :type tag: str, optional
    :param dummy: If the Index set is a dummy set, elements are created immediately. Defaults to False.
    :type dummy: bool, optional

    :ivar _: Elements of the index set
    :vartype _: list[X]
    :ivar tag: Tag/details
    :vartype tag: str
    :ivar ordered: Ordered set, True if size is given
    :vartype ordered: bool
    :ivar name: Name, set by the program
    :vartype name: str
    :ivar n: Number id, set by the program
    :vartype n: int
    :ivar ltx: LaTeX representation
    :vartype ltx: str

    :raises ValueError: If both members and size are given
    :raises ValueError: If indices of elements (P, V) are not compatible
    :raises ValueError: If index set is not ordered and step is given

    Examples
    --------
    .. code-block:: python

        p = Program()
        p.s1 = I('a', 'b', 'c')
        p.s2 = I('a', 'd', 'e', 'f')
        
        # Intersection
        p.s1 & p.s2
        # I('a')
        
        # Union
        p.s1 | p.s2
        # I('a', 'b', 'c', 'd', 'e', 'f')
        
        # Symmetric difference
        p.s1 ^ p.s2
        # I('b', 'c', 'd', 'e', 'f')
        
        # Difference
        p.s1 - p.s2
        # I('b', 'c')
    """

    def __init__(
        self,
        *members: str | int,
        size: int = None,
        mutable: bool = False,
        tag: str = None,
        ltx: str = None,
        dummy: bool = False,
    ):
        self.tag = tag
        self.mutable = mutable
        # set by program
        self.name = ''
        self.n = None

        # this is when children single element sets are created
        # These will be set in ._
        self.parent: list[Self] = []
        self.pos: list[int] = []
        self._: list[Self] = []

        # if it is a slice
        self.slice: slice = None

        # hash passed on along with name
        # self._hash = ''

        if size:
            if members:
                raise ValueError(
                    'An index set can either be defined by members or size, not both'
                )
            self.size = size
            self.members = []
            self.ordered = True

        elif members:
            self.size = len(members)
            self.members = members
            self.ordered = False

        else:
            self._ = []
            self.members = []
            self.ordered = False

        # # compound sets collect a list of children sets
        # # from which they are made
        # self.children: list[Self] = []
        # # These are used for index arrays for function (F) sets
        # self.one: I = None
        # self.two: I = None

        self.parameters = []
        self.variables = []
        self.functions = []
        self.constraints = []

        # if latex name is given
        self.ltx = ltx

        if dummy:
            self.case = ICase.DUMMY
            self.birth_elements()

        else:
            self.case: ICase = None

    def birth_elements(self):
        """Create elements for the index set"""
        # if self.ordered:
        self.size = int(self.size)
        for n in range(self.size):
            # this is called from outside
            # (once the name is set)
            # for an ordered index set
            # create new index
            index = I()
            # append parent
            index.parent.append(self)
            # update position in parent
            index.pos.append(n)
            # set that this is ordered
            index.ordered = True
            # give the name
            index.name = rf'{self}[{n}]'
            index._hash = hash(index.name)
            # the only element in element (index set of size one)
            # is itself
            index._ = [index]
            index.size = 1
            index.members = [index.name]
            self._.append(index)
            index.ltx = rf'{self.ltx}[{n}]'

    # -----------------------------------------------------
    #                    Modifiers
    # -----------------------------------------------------

    def step(self, n: int) -> list[Self]:
        """Step up or down the index set
        Args:
            i (int): Step size
        """
        if not self.ordered:
            raise ValueError(
                'Index set is not ordered, cannot step up or down the index set'
            )
        if not n:
            # if no step (0)
            return self

        # else create a new index set
        index = I()
        # if step is negative
        if n < 0:
            in_index = self._[:n]
            index._ = [None] * -n + in_index
            # the negative sign will come with n
            index.name = f'{self.name}{n}'
        else:
            in_index = self._[n:]
            index._ = in_index + [None] * n
            # + needs to be provided
            index.name = f'{self.name}+{n}'

        # update the members
        index.members = [i.name for i in in_index]
        # note that this is a subset of self
        index.parent = self
        # the size is still the same
        index.size = self.size
        # only done for index set
        index.ordered = True
        # latex representation
        index.ltx = self.ltx

        return index

    # -----------------------------------------------------
    #                    Printing
    # -----------------------------------------------------

    # def nsplit(self):
    #     """Split the name
    #     If there is an underscore, the name is split into name and superscript
    #     """
    #     if '_' in self.name:
    #         name, sup = self.name.split('_')
    #         if sup:
    #             return r'{' + name + r'}', r'^{' + sup + r'}'
    #         # this is used for negation sometimes
    #         return '-' + self.name[:-1], ''
    #     return r'{' + self.name + r'}'

    def latex(self, descriptive: bool = True, int_not: bool = False) -> str:
        """LaTeX representation
        Args:
            descriptive (bool): print members of the index set
            int_not (bool): Whether to display the set in integer notation.
        """

        if not self.name:
            return ''

        # if the name has underscores, replace them with \_
        ltx = self.name.replace('_', r'\_')

        if self.parent and any(parent.ordered for parent in self.parent):
            ltx = ltx.replace('[', '_{').replace(']', '}')
            ltx = r'{' + ltx + r'}'

        else:
            ltx = ltx.replace('[', '{').replace(']', '}')

        # name, sup = self.nsplit()
        ltx = ltx.replace('|', r'\cup')
        # mathcal = rf'\mathcal{{{name}{sup}}}'

        if self.parent:
            return ltx

        if self.case == ICase.SELF:
            # if this is a self contained index
            return ''

        if descriptive:
            if self.ordered:
                if int_not:
                    return (
                        rf'\{{ i = \mathbb{{{ltx}}} \mid '
                        rf'{self._[0]} \leq i \leq {self._[-1]} \}}'
                    )
                members = (
                    r', '.join(x.latex() for x in self._)
                    if len(self) < 5
                    else rf'{self._[0].latex()},..,{self._[-1].latex()}'
                )
                return rf'{ltx} = \{{ {members} \}}'

            members = r', '.join(x.latex() for x in self._)
            return rf'{ltx} = \{{ {members} \}}'

        return ltx

    def show(self, descriptive: bool = True):
        """Display the set

        Args:
            descriptive (bool, optional): Displays all members in the index set. Defaults to False.
        """
        display(Math(self.latex(descriptive)))

    def mps(self, pos: int) -> str:
        """MPS representation
        Args:
            pos (int): Position of the member in the set
        """
        return rf'_{self[pos]}'.upper()

    def lp(self, pos: int) -> str:
        """LP representation
        Args:
            pos (int): Position of the member in the set
        """
        return rf'_{self[pos]}'

    # -----------------------------------------------------
    #                    Birth
    # -----------------------------------------------------

    def birth_index(self, name: str, members: list[Self]) -> Self:
        """Updates the parent, sets new positions and mutable/ordered attributes"""
        # set new members for the index
        index = I()
        # set a name for the new index
        index.name = name
        # update the members of the index
        # doing this from outside avoids creating
        # element index sets (X) again
        index.members = members
        index.ordered = self.ordered
        return index

    # -----------------------------------------------------
    #                    Operators
    # -----------------------------------------------------

    # Avoid running instance checks
    def __eq__(self, other: Self):
        # equality checks for index sets are only done
        # on the basis of names
        return is_(self, other)

    def __and__(self, other: Self):
        # Members that exist in both Index sets
        _and = [i for i in self.members if i in other.members]

        return self.birth_index(rf'{self.name} & {other.name}', _and)

    def __or__(self, other: Self):
        # members that exist in either self or other
        # make a copy of the members in self
        _or = list(self.members)
        # if a member in other is not included
        for i in other.members:
            if not i in _or:
                # append it to the list
                _or.append(i)
        # mutable sets will have the same name
        # and are mutated using | (__or__)
        # repeated names are not allowed,
        # thus if the same name is coming in,
        # the index is definitely being mutated
        if self.name == other.name:
            return self.birth_index(self.name, _or)

        # else create a new name that reflects the operation
        return self.birth_index(rf'{self.name} | {other.name}', _or)

    def __xor__(self, other: Self):
        # members that exist in either self or other, but not both
        # create an empty list to be updated
        _xor: list[Self] = []
        # if something is in self but not in other
        for i in self.members:
            if not i in other.members:
                _xor.append(i)
        # if something is in other but not in self
        for i in other.members:
            if not i in self.members:
                _xor.append(i)
        return self.birth_index(rf'{self.name} ^ {other.name}', _xor)

    def __sub__(self, other: int | Self):
        # other is an integer, step down the index set
        if isinstance(other, int):
            if len(self) == 1:
                return
            return self.step(-other)

        # members from other are removed from self
        # if other is some type of an Index set
        # create an empty list
        _sub = []
        # if a member of self is in the other set
        for i in self.members:
            if not i in other.members:
                # do not append
                _sub.append(i)
        return self.birth_index(rf'{self.name} - {other.name}', _sub)

    def __add__(self, other: int | Self):
        # if other is an integer, step up the index set
        if isinstance(other, int):
            return self.step(other)

        raise NotImplementedError(
            'Addition of Index sets is not implemented. Use | or the "or" operator for union.\n'
            '+  can be used to step up the index set by an integer.'
        )

    def __mul__(self, other: Self | tuple | None):
        # product of two Index sets

        if other is None:
            # This will likely be used mostly for element indices
            # allowing indices to be skipped while generating elements
            return None

        # if other is a tuple, return a tuple with self as the first element
        if isinstance(other, tuple):
            return (self,) + other

        # if other is an Index set, return a tuple of index sets
        return (self, other)

    def __rmul__(self, other: Self | tuple | None):
        if other is None:
            return None

        if isinstance(other, int):
            # This allows the use of math.prod
            if other == 1:
                return self

        # the only other allowed instance for which this
        # is called is tuple
        # Not running an instance check to save time
        return other + (self,)

    # -----------------------------------------------------
    #                    Vector
    # -----------------------------------------------------

    def __len__(self) -> int:
        return len(self._)
        # return len([i for i in self._ if not isinstance(i, Skip)])

    def __iter__(self) -> Self:
        return iter(self._)

    def __getitem__(self, key: int | str | slice) -> Self:
        if isinstance(key, slice):
            #  if this is a slice [start:stop]
            # generate a new index set for that stretch
            index = I(*self._[key], mutable=self.mutable, tag=self.tag)
            # mark this as a subset of self
            index.parent = self
            index.slice = key
            # note the start and stops
            if key.start is None:
                index.name = rf'{self.name}[0:{key.stop}]'
            else:
                index.name = rf'{self.name}[{key.start}:{key.stop}]'

            index.ordered = self.ordered
            index._ = self._[key]
            return index
        return self._[key]

    def __contains__(self, other: Self):
        return True if other in self._ else False

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    # __str__ = lambda self: self.name
    # __repr__ = __str__
    # __hash__ = lambda self: hash(self.name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

    def __hash__(self):
        # this seems to work the fastest
        # not sure how pythonic this is though
        try:
            return self._hash

        except AttributeError:
            self._hash = hash(self.name)
            # self._hash = hash(self.name)
            return self.__hash__()

        # if not self._hash:
        #     self._hash = hash(self.name)

        # return self._hash

    # -----------------------------------------------------
    #                    Export
    # -----------------------------------------------------

    def sympy(self):
        """Sympy representation"""
        if has_sympy:
            return FiniteSet(*[str(s) for s in self._])
        print(
            "sympy is an optional dependency, pip install gana[all] to get optional dependencies"
        )

    def pyomo(self):
        """Pyomo representation"""
        if has_pyomo:
            if self.ordered:
                return PyoRangeSet(i, len(self), doc=self.tag)

            return PyoSet(initialize=[i.name for i in self._], doc=self.tag)
        print(
            "pyomo is an optional dependency, pip install gana[all] to get optional dependencies"
        )



"""Program"""

import warnings
from dataclasses import dataclass, field

from numpy import array as nparray
from numpy import zeros as npzeros
from pandas import DataFrame
from ppopt.mplp_program import MPLP_Program

from ..operators.composition import inf, sup
from ..sets.cases import Elem, ICase, PCase
from ..sets.constraint import C
from ..sets.function import F as Func
from ..sets.index import I
from ..sets.objective import O
from ..sets.parameter import P
from ..sets.theta import T
from ..sets.variable import V
from .solution import Solution

# optional dependencies
try:
    from gurobipy import read as gpread

    has_gurobi = True
except ImportError:
    has_gurobi = False

try:
    from IPython.display import Markdown, display

    has_ipython = True
except ImportError:
    has_ipython = False

try:
    from pyomo.environ import ConcreteModel as PyoModel

    has_pyomo = True
except ImportError:
    has_pyomo = False


try:

    has_pandas = True
except ImportError:
    has_pandas = False


@dataclass
class Prg:
    """
    A mathematical program.

    Can be a linear (LP), integer (IP), or mixed-integer (MIP).

    :param name: Name of the program. Defaults to 'prog'.
    :type name: str, optional
    :param tol: Tolerance. Defaults to None.
    :type tol: float, optional
    :param canonical: Whether to use canonical form. Defaults to True.
    :type canonical: bool, optional
    :param tag: Tag for the program. Defaults to ''.
    :type tag: str, optional

    :ivar names: Names of declared sets
    :vartype names: list[str]
    :ivar sets: Object to hold set objects
    :vartype sets: Sets
    :ivar names_idx: Names of the index elements
    :vartype names_idx: list[str]
    :ivar indices: Index sets
    :vartype indices: list[X]
    :ivar variables: Variable sets
    :vartype variables: list[Var]
    :ivar thetas: Parametric variable sets
    :vartype thetas: list[PVar]
    :ivar functions: Function sets
    :vartype functions: list[Func]
    :ivar constraints: Constraint sets
    :vartype constraints: list[Cons]
    :ivar objectives: Objective sets
    :vartype objectives: list[Obj]

    :raises ValueError: If overwriting a set
    """

    name: str = field(default='prog')
    tol: float = field(default=None)
    canonical: bool = field(default=True)
    tag: str = field(default='')

    def __post_init__(self):

        # ------ collections --------
        # index (I)
        self.index_sets: list[I] = []
        self.indices: list[I] = []

        # variable (V)
        self.variable_sets: list[V] = []
        self.variables: list[V] = []

        # parameter (P)
        self.parameter_sets: list[P] = []  # parameter sets
        # parameter set elements are just numeric

        # parametric variable (T)
        self.theta_sets: list[T] = []
        self.thetas: list[T] = []

        # function (F)
        self.function_sets: list[Func] = []
        self.functions: list[Func] = []

        # constraint (C)
        self.constraint_sets: list[C] = []
        self.constraints: list[C] = []

        self.categories_sets: dict[str, list[C]] = {}  # categories of constraint sets
        self.categories: dict[str, list[C]] = {}  # categories of constraints

        self.fcategories_sets: dict[str, list[Func]] = {}  # categories of function sets
        self.fcategories: dict[str, list[Func]] = {}  # categories of functions

        # objective (O)
        self.objectives: list[O] = []

        # ------ names --------
        self.names: list[str] = []  # element names

        self.names_index_sets: list[str] = []  # index sets
        self.names_indices: list[str] = []  # elements
        self.names_variable_sets: list[str] = []  # variable sets
        self.names_parameter_sets: list[str] = []  # parameter sets
        self.names_theta_sets: list[str] = []  # parametric variable sets
        self.names_function_sets: list[str] = []  # function sets
        self.names_constraint_sets: list[str] = []  # constraints
        self.names_objectives: list[str] = []  # objectives

        # ------ counts --------

        # index (I)
        self.n_index_sets: int = 0  # index sets
        self.n_index_elements: int = 0  # elements

        # variable (V)
        self.n_variable_sets: int = 0
        self.n_variables: int = 0

        # parameter (P)
        self.n_parameter_sets: int = 0

        # parametric variable (T)
        self.n_theta_sets: int = 0
        self.n_thetas: int = 0

        # function (F)
        self.n_function_sets: int = 0
        self.n_functions: int = 0

        # constraint (C)
        self.n_constraint_sets: int = 0
        self.n_constraints: int = 0

        # objective (O)
        self.n_objectives: int = 0  # objectives

        # flag for is optimized
        self.optimized = False

        # the solution object
        self.solution: dict[int, Solution] = {}

        # number of solutions
        self.n_sol: int = 0

        # solution matrix

        self.X: dict[int, list[float | int]] = {}

    def add_index(self, name: str, index: I):
        """Adds new index to program

        Args:
            name (str): name of index
            index (I): index set to be added
        """
        self.names.append(name)
        self.names_index_sets.append(name)
        # give the index a name
        index.name = name
        index._hash = hash(name)
        # This is the nth index set (0 indexed)
        index.n = self.n_index_sets
        # update the number of index sets
        self.n_index_sets += 1
        # update the list of indices
        self.index_sets.append(index)

    def add_indices(self, index: I, members: list[str] = None):
        """Adds indices from an index set to the program

        index (I): Index set who elements are to be added to the program.
        members (list[str]): List of members to be added to the index set.
        """

        if not members:
            # if specific members are not specified
            # add all of them as elements
            members = index.members

        # for unordered sets, the elements are set on the program
        for n, member in enumerate(members):

            if not member in self.names_indices:
                # if this element has not been added to the program before
                # create an element set
                element = I()
                # set the name
                element.name = member
                element._hash = hash(member)
                # this is the nth element (0 indexed)
                element.n = self.n_index_elements
                # update the number of elements
                self.n_index_elements += 1
                # this is a new element, so set the new name
                self.names_indices.append(member)
                self.indices.append(element)
                # if new element, it should be set on the program
                _new_elm = True

            else:

                # if the element is already set on the program
                # it can be part of another set already, get it from the program
                # e.g. human is both part of the kingdom index set animalia
                # and the family set hominidae
                # so its parent sets need to be animalia as well as hominidae
                # so update parent and add position in the new index set
                element: I = getattr(self, member)
                # these should not be set again
                _new_elm = False

            # update the index (I) as a parent
            element.parent.append(index)
            # update the position of the element in the index set
            element.pos.append(len(index._))
            # These are neither ordered or unordered sets
            element.ordered = None
            # element has only one member, itself
            element._ = [element]
            # now update the element in the index set
            index._.append(element)

            if _new_elm:
                # if this is a new element, set in
                setattr(self, member, element)

    def add_variable(self, name: str, variable: V):
        """Adds new variable set to program

        Args:
            name (str): name of variable set
            variable (V): variable set to be added
        """
        self.names.append(name)
        self.names_variable_sets.append(name)
        # give the variable a name
        variable.name = name
        # This is the nth variable set (0 indexed)
        variable.n = self.n_variable_sets
        # update the number of variable sets
        self.n_variable_sets += 1
        # update the list of variables
        self.variable_sets.append(variable)

        variable.birth_variables(n_start=self.n_variables)
        # update the list of variables
        self.variables.extend(variable._)
        self.n_variables += len(variable._)

    def mutate_variable(self, variable_ex: V, variable_new: V):
        """Mutates an existing variable set in the program
        Args:
            variable_ex (V): exisiting variable set to be mutated
            variable (V): incoming variable set to be added
        """
        # birth the new variables
        # inform that this is a mutation
        variable_new.birth_variables(mutating=True, n_start=self.n_variables)

        # the positions need to be pushed ahead
        pos_start = len(variable_ex)
        # note: if a variable already exists in the existing variable set
        # then, it is not added.
        # thus the position (and hence name) depends on the existing variables
        # this keeps a count of the number of variables added
        n = 0
        _name = variable_ex.name  # name of the existing variable set
        # iterate through all the new variables
        # update a list of new variable elements to be added
        var_add: list[V] = []
        for idx, v in variable_new.map.items():
            if idx is None:
                # for a None index, skip
                continue
            # only update exisitng variable
            # if not already in the existing variable set
            if not idx in variable_ex.map:
                # set the position of the new variable
                _pos = pos_start + n
                v.pos = _pos
                # set the name based on position in existing variable set
                v.name = f'{_name}[{_pos}]'  # give a name
                # set the parent to the existing variable set
                v.parent = variable_ex
                # update the variable map
                variable_ex.map[idx] = v
                # update the variable set
                variable_ex._.append(v)
                # and the added variables
                var_add.append(v)
                # update the iter counter
                n += 1
        if n > 0:
            variable_ex.n_splices += 1
            # update the existing variable index
            # only if something new has been added
            var_ex_idx = tuple(
                [i[0] if isinstance(i, list) else i for i in variable_ex.index]
            )
            var_new_idx = tuple(
                [i[0] if isinstance(i, list) else i for i in variable_new.index]
            )

            if variable_ex.n_splices > 2:
                # if there are more than 2 splices
                variable_ex.index = {*variable_ex.index, var_new_idx}
            else:
                variable_ex.index = {var_ex_idx, var_new_idx}
            self.n_variables += n
            self.variables.extend(var_add)

    def add_parameter(self, name: str, parameter: P):
        """Adds new parameter set to program

        Args:
            name (str): name of parameter set
        parameter (P): parameter set to be added
        """
        self.names.append(name)
        self.names_parameter_sets.append(name)
        # give the parameter a name
        if not parameter.name:
            parameter.name = name
        # This is the nth parameter set (0 indexed)
        parameter.n = self.n_variable_sets
        # update the number of parameter sets
        self.n_parameter_sets += 1
        # update the list of parameters
        self.parameter_sets.append(parameter)

    def mutate_parameter(self, parameter_ex: P, parameter_new: P):
        """Mutates an existing parameter set in the program
        Args:
            parameter_ex (P): exisiting parameter set to be mutated
            parameter_new (P): incoming parameter set to be added
        """
        n = 0  # count of parameters added to set
        for idx, p in parameter_new.map.items():

            if idx is None:
                # for a None index, skip
                continue

            if idx in parameter_ex.map:
                # warn if number is being replaced
                warnings.warn(
                    f'The value{parameter_ex.map[idx]} is being overwritten by {p} at index {idx}'
                )

            # set the position of the new parameter
            parameter_ex.map[idx] = p
            parameter_ex._.append(p)
            n += 1

        if n > 0:
            parameter_ex.n_splices += 1
            # update the existing parameter index
            # only if something new has been added
            if parameter_ex.n_splices > 2:
                parameter_ex.index = {*parameter_ex.index, parameter_new.index}
            else:
                parameter_ex.index = {parameter_ex.index, parameter_new.index}

    def add_theta(self, name: str, theta: T):
        """Adds new theta set to program

        Args:
            name (str): name of theta set
            theta (V): theta set to be added
        """
        self.names.append(name)
        self.names_theta_sets.append(name)
        # give the theta a name
        theta.name = name
        # This is the nth theta set (0 indexed)
        theta.n = self.n_theta_sets
        # update the number of theta sets
        self.n_theta_sets += 1
        # update the list of thetas
        self.theta_sets.append(theta)

        theta.birth_thetas(n_start=self.n_thetas)
        # update the list of thetas
        self.thetas.extend(theta._)
        self.n_thetas += len(theta._)

    def update_theta(self, constraint: C):
        """Updates the theta set and thetas in a constraint

        Args:
            constraint (C): constraint with thetas to be updated
        """
        for n, theta in enumerate(constraint.function.rhs_thetas):
            self.add_theta(name=f'θ{self.n_theta_sets}', theta=theta)
            # the last theta added is the one just made
            theta = self.theta_sets[-1]
            # replace the theta in the constraint rhs list
            constraint.function.rhs_thetas[n] = theta
            # Create the F and Z matrices for the constraint
            # this only handles addition or subtraction with T
            if (
                constraint.function.one_type == Elem.F
                and constraint.function.one.two_type == Elem.T
            ):
                # this is of the form V/F +- Th1 +- Th2
                if constraint.function.add:
                    constraint.function.F = [
                        i + [-1] for i in constraint.function.one.F
                    ]
                    constraint.function.Z = [
                        i + [theta._[pos].n]
                        for pos, i in enumerate(constraint.function.one.Z)
                    ]
                if constraint.function.sub:
                    constraint.function.F = [i + [1] for i in constraint.function.one.F]
                    constraint.function.Z = [
                        i + [theta.n] for i in constraint.function.one.Z
                    ]
                    constraint.function.Z = [
                        i + [theta._[pos].n]
                        for pos, i in enumerate(constraint.function.one.Z)
                    ]
            else:

                # this is of the form V/F +- Th1
                if constraint.function.add:
                    constraint.function.F = [[-1]] * len(constraint.one)
                    constraint.function.Z = [
                        [theta._[pos].n] for pos in range(len(constraint.one))
                    ]
                if constraint.function.sub:
                    constraint.function.F = [[1]] * len(constraint.one)
                    constraint.function.Z = [
                        [theta._[pos].n] for pos in range(len(constraint.one))
                    ]

            for n, cons in enumerate(constraint._):
                # update the constraint element matrices
                cons.function.F = constraint.function.F[n]
                cons.function.Z = constraint.function.Z[n]
                # update two
                cons.function.two = theta[n]
                cons.function.give_name()

    def mutate_theta(self, theta_ex: T, theta_new: T):
        """Mutates an existing theta set in the program
        Args:
            theta_ex (T): exisiting theta set to be mutated
            theta (T): incoming theta set to be added
        """
        # birth the new thetas
        # inform that this is a mutation
        theta_new.birth_thetas(mutating=True, n_start=self.n_thetas)

        # the positions need to be pushed ahead
        pos_start = len(theta_ex)
        # note: if a theta already exists in the existing theta set
        # then, it is not added.
        # thus the position (and hence name) depends on the existing thetas
        # this keeps a count of the number of thetas added
        n = 0
        _name = theta_ex.name  # name of the existing theta set
        # iterate through all the new thetas
        # update a list of new theta elements to be added
        tht_add: list[T] = []
        for idx, t in theta_new.map.items():
            if idx is None:
                # for a None index, skip
                continue
            # only update exisitng theta
            # if not already in the existing theta set
            if not idx in theta_ex.map:
                # set the position of the new theta
                _pos = pos_start + n
                t.pos = _pos
                # set the name based on position in existing theta set
                t.name = f'{_name}[{_pos}]'  # give a name
                # set the parent to the existing theta set
                t.parent = theta_ex
                # update the theta map
                theta_ex.map[idx] = t
                # update the theta set
                theta_ex._.append(t)
                # and the added thetas
                tht_add.append(t)
                # update the iter counter
                n += 1
        if n > 0:
            # update the existing theta index
            # only if something new has been added
            tht_ex_idx = tuple(
                [i[0] if isinstance(i, list) else i for i in theta_ex.index]
            )
            tht_new_idx = tuple(
                [i[0] if isinstance(i, list) else i for i in theta_new.index]
            )
            theta_ex.index = {tht_ex_idx, tht_new_idx}
            self.n_thetas += n
            self.thetas.extend(tht_add)

    def add_function(self, name: str, function: Func):
        """Add a function set to the program

        Args:
            name (str): name of function
            function (F): function object
        """
        self.names.append(name)
        self.names_function_sets.append(name)
        # give the function a name
        # but do not add it to name
        # we want the hash to the equation
        # but the attribute name to be name
        function.pname = name
        # This is the nth function set (0 indexed)
        function.n = self.n_function_sets
        # update the number of function sets
        self.n_function_sets += 1
        # update the list of functions
        self.function_sets.append(function)

        # # update the list of functions
        self.functions.extend(function._)
        for n, f in enumerate(function._):
            # this is the nth function declared
            f.n = self.n_functions + n

        # # update the number of functions
        self.n_functions += len(function._)

    def replace_function(self, function_ex: Func, function_new: Func):
        """Replaces an existing function set in the program

        Args:
            function_ex (F): existing function set to be mutated
            function_new (F): new function set to replace the existing one
        """

        # just replace the existing function set
        # take the old constraints number and pname
        function_new.n = function_ex.n

        function_new.pname = function_ex.pname
        # replace the function set in the program
        self.function_sets[self.function_sets.index(function_ex)] = function_new
        # update the list of functions

        # first, number the functions in the new set
        for n, f in enumerate(function_new._):
            f.n = self.n_functions + n

        # second, remove the old functions
        self.functions = (
            self.functions[: function_ex._[0].n]
            + self.functions[function_ex._[-1].n + 1 :]
        )

        # third, add the new functions
        self.functions.extend(function_new._)

        # four, clean up the features_in list for variables in old function
        for v in function_ex.variables:
            if function_ex in v.min_by:
                v.min_by.remove(function_ex)

            for var in v._:
                for func in function_ex._:
                    if func in var.min_by:
                        _min_by = var.min_by
                        _min_by.remove(func)
                        var.min_by = _min_by

    def add_constraint(self, name: str, constraint: C):
        """Adds a constraint set to the program

        Args:
            name (str): name of constraint
            constraint (C): constraint object
        """
        self.names.append(name)
        self.names_constraint_sets.append(name)
        # give the constraint a name
        # but do not add it to name
        # we want the hash to the equation
        # but the attribute name to be name
        constraint.pname = name
        # This is the nth constraint set (0 indexed)
        constraint.n = self.n_constraint_sets
        # update the number of constraint sets
        self.n_constraint_sets += 1
        # update the list of constraints
        self.constraint_sets.append(constraint)

        constraint.update_variables()

        # update the list of constraints
        self.constraints.extend(constraint._)
        for n, c in enumerate(constraint._):
            # this is the nth constraint declared
            c.n = self.n_constraints + n

        # update the number of constraints
        self.n_constraints += len(constraint._)

        if constraint.function.rhs_thetas:
            # if the constraint has thetas in them
            # then update the thetas in the function
            self.update_theta(constraint)

    def replace_constraint(self, constraint_ex: C, constraint_new: C):
        """Replaces an existing constraint set in the program

        Args:
            constraint_ex (C): existing constraint set to be mutated
            constraint_new (C): new constraint set to replace the existing one
        """
        # just replace the existing constraint set
        # take the old constraints number and pname
        constraint_new.n = constraint_ex.n
        constraint_new.pname = constraint_ex.pname
        if not constraint_ex.category == 'General':

            constraint_new.categorize(constraint_ex.category)

        # replace the constraint set in the program
        # self.constraint_sets[self.constraint_sets.index(constraint_ex)] = constraint_new
        self.constraint_sets[constraint_ex.n] = constraint_new

        # replace the constraint in the program
        # let new constraint take n of the old constraint
        for cons_new, cons_ex in zip(constraint_new._, constraint_ex._):

            # self.constraints[self.constraints.index(cons_ex)] = cons_new
            self.constraints[cons_ex.n] = cons_new

            cons_new.n = cons_ex.n

        for cons_ex, cons_new in zip(constraint_ex._, constraint_new._):
            cons_new.cons_by_pos = cons_ex.cons_by_pos
            for v in cons_ex.variables:
                v.cons_by[cons_ex.cons_by_pos[v]] = cons_new

            for v in cons_new.variables:
                if cons_new not in v.cons_by:
                    cons_new.cons_by_pos[v] = len(v.cons_by)
                    v.cons_by.append(cons_new)

        # for variable in self.variable_sets:

        #     if constraint_ex in variable.cons_by:
        #         # if variable is constrained by the old constraint
        #         # remove the constraint from the variable
        #         variable.cons_by[variable.cons_by.index(constraint_ex)] = constraint_new

        #     for var in variable._:

        #         for cons_new, cons_ex in zip(constraint_new._, constraint_ex._):
        #             # if variable is constrained by the old constraint
        #             # update the cons_by list
        #             if cons_ex in var.cons_by:
        #                 var.cons_by[var.cons_by.index(cons_ex)] = cons_new

        # for cons_new in constraint_new._:
        #     # there could be new variables in the updated constraint
        #     for var in cons_new.variables:

        #         # if the variable is in the new constraint
        #         if cons_new not in var.cons_by:

        #             # update the cons_by list
        #             var.cons_by.append(cons_new)

    def add_objective(self, objective: O):
        """Adds an objective set to the program

        Args:
            objective (O): objective object
        """
        self.names.append(objective.pname)
        self.names_objectives.append(objective.pname)
        # This is the nth objective set (0 indexed)
        objective.n = self.n_objectives
        # update the number of objective sets
        self.n_objectives += 1
        # update the list of objectives
        self.objectives.append(objective)
        self.function_sets.append(objective.function)
        objective.update_variables()

    def __setattr__(self, name, value) -> None:

        _mutation = False  # skip setting set

        if isinstance(value, I):

            if not name in self.names_index_sets and not name in self.names_indices:

                if len(value.members) == 1 and value.members[0] == name:
                    # There is a special case, where a self contained set is passed
                    # in that case, add the index
                    self.add_index(name, value)
                    # but the only set it contains is itself
                    value._ = [value]
                    value.case = ICase.SELF

                else:

                    # check if index already exists in the program
                    # or is the name of an element
                    # element are set by their parents
                    # if this is a new object, it can be safely added
                    self.add_index(name, value)

                    if value.ordered:
                        # for ordered set, the elements are not set on the program
                        value.birth_elements()

                    else:
                        self.add_indices(value)

            elif not name in self.names_indices:
                # the set could be already declared, and mutable
                # is being declared as part of another index set
                # in which case get the original set to update
                index_ex: I = getattr(self, name)  # existing index set

                # if not mutable, raise error
                if not index_ex.mutable:
                    raise ValueError(
                        f'{self.name}: Overwriting index {name}. Set mutable=True if index needs to be updated'
                    )

                # if an index is being mutated, skip setting
                _mutation = True

                # # for collections, each element is a set of size 1
                if not value.ordered:
                    # update the exisiting index set with the new elements
                    _members = []
                    for member in value.members:
                        if member not in index_ex.members:
                            # if the member is not already in the index set
                            _members.append(member)

                    self.add_indices(index_ex, _members)

        elif isinstance(value, V):

            if not name in self.names_variable_sets:
                if not value.name in self.names_variable_sets:
                    # if variable set is new, add it to the program
                    # another check we do, is if variable is being added to the program
                    # but the variable already exists
                    # this happens in a case such as this:
                    # p.f0 = p.v - 3
                    # p.f1 = p.f0 + 3
                    # This is returning a variable set that already exists in the program
                    if not any(None for _ in value.index):
                        self.add_variable(name, value)

            else:
                variable_ex: V = getattr(self, name)  # existing variable set

                # if not mutable, raise error
                if not variable_ex.mutable:
                    raise ValueError(
                        f'{self.name}: Overwriting variable {name}. Set mutable=True if variable needs to be updated'
                    )

                # if an index is being mutated, skip setting
                _mutation = True

                # give a name because
                # the incoming variable set will birth variables too
                # and they will need a name
                value.name = name
                self.mutate_variable(variable_ex, value)

        elif isinstance(value, P):
            if not name in self.names_parameter_sets:
                # if parameter is new, add it to the program

                if value.case in [PCase.NEGSET, PCase.SET]:
                    self.add_parameter(name, value)

            else:
                parameter_ex: P = getattr(self, name)  # existing parameter set

                # if not mutable, raise error
                if not parameter_ex.mutable:
                    raise ValueError(
                        f'{self.name}: Overwriting parameter {name}. Set mutable=True if parameter needs to be updated'
                    )

                # if an index is being mutated, skip setting
                _mutation = True
                # mutate the parameter
                self.mutate_parameter(parameter_ex, value)

        elif isinstance(value, T):

            if not name in self.names_theta_sets:
                self.add_theta(name, value)

            else:
                theta_ex: T = getattr(self, name)
                if not theta_ex.mutable:
                    raise ValueError(
                        f'{self.name}: Overwriting theta {name}. Set mutable=True if theta needs to be updated'
                    )
                _mutation = True
                self.mutate_theta(theta_ex, value)

        elif isinstance(value, Func):
            if not name in self.names_function_sets:
                self.add_function(name, value)
            else:
                # if function is being mutated
                # replace existing function with the new one
                self.replace_function(getattr(self, name), value)
                # but still set it

        elif isinstance(value, C):

            if not name in self.names_constraint_sets:
                self.add_constraint(name, value)

            else:
                self.replace_constraint(getattr(self, name), value)

        elif isinstance(value, O):
            self.add_objective(value)

        if not _mutation:
            super().__setattr__(name, value)

    # --------------------------------------------------
    #               Subsets
    # --------------------------------------------------

    @property
    def nncons_sets(self) -> list[C]:
        """non-negativity constraint sets"""
        return [x for x in self.constraint_sets if x.nn]

    @property
    def eqcons_sets(self) -> list[C]:
        """equality constraint sets"""
        return [x for x in self.constraint_sets if not x.leq]

    @property
    def leqcons_sets(self) -> list[C]:
        """less than or equal constraint sets"""
        return [x for x in self.constraint_sets if x.leq and not x.nn]

    def nncons(self, n: bool = False) -> list[int | C]:
        """non-negativity constraints"""
        if n:
            return [x.n for x in self.constraints if x.nn]
        return [x for x in self.constraints if x.nn]

    def eqcons(self, n: bool = False) -> list[int | C]:
        """equality constraints"""
        if n:
            return [x.n for x in self.constraints if not x.leq]
        return [x for x in self.constraints if not x.leq]

    def leqcons(self, n: bool = False) -> list[int | C]:
        """less than or equal constraints"""
        if n:
            return [x.n for x in self.constraints if x.leq and not x.nn]
        return [x for x in self.constraints if x.leq and not x.nn]

    def cons(self, n: bool = False) -> list[int | C]:
        """constraints"""
        return self.leqcons(n) + self.eqcons(n) + self.nncons(n)

    def nnvars(self, n: bool = False) -> list[int | V]:
        """non-negative variables"""
        if n:
            return [x.n for x in self.variables if x.nn]
        return [x for x in self.variables if x.nn]

    def bnrvars(self, n: bool = False) -> list[int | V]:
        """binary variables"""
        if n:
            return [x.n for x in self.variables if x.bnr]
        return [x for x in self.variables if x.bnr]

    def intvars(self, n: bool = False) -> list[int | V]:
        """integer variables"""
        if n:
            return [x.n for x in self.variables if x.itg]
        return [x for x in self.variables if x.itg]

    def contvars(self, n: bool = False) -> list[int | V]:
        """continuous variables"""
        if n:
            return [x.n for x in self.variables if not x.bnr and not x.itg]
        return [x for x in self.variables if not x.bnr and not x.itg]

    def renumber(self):
        """Renumbers the constraints, just to be sure"""
        for n, c in enumerate(self.cons()):
            c.n = n

    # --------------------------------------------------
    #               Matrices
    # --------------------------------------------------

    # DONOT call these for large programs
    # Going to run into memory issues
    # TODO: make sparse matrix options

    @property
    def B(self) -> list[float]:
        """RHS Parameter vector"""
        return [c.B for c in self.cons()]

    @property
    def A(self) -> list[list[float]]:
        """Matrix of Variable coefficients"""
        constraints = self.cons()
        _A = []
        for _ in constraints:
            row = [0] * len(self.variables)
            _A.append(row)

        for n, c in enumerate(constraints):
            for x, a in zip(c.P, c.A):
                _A[n][x] = a
        return _A

    @property
    def F(self) -> list[list[float]]:
        """Matrix of Parameteric Variable coefficients"""
        constraints = self.cons()
        _F = []
        for _ in constraints:
            row = [0] * len(self.theta_sets)
            _F.append(row)

        n = 0
        for c in constraints:
            m = 0
            for z, f in zip(c.Z, c.F):
                if z and z[m] is not None:
                    _F[n][z[m]] = f[m]
                m += 1
                n += 1
        return _F

    @property
    def C(self) -> list[float]:
        """Objective Coefficients"""
        # no objectives have been set
        if len(self.objectives) == 0:
            return []

        if len(self.objectives) == 1:
            # only one objective has been set
            obj = self.objectives[0]

            _C = [0] * len(self.variables)  # initialize with zeros
            for n, v in enumerate(obj.variables):
                _C[obj.P[n]] = obj.C[n]

        return _C

        # TODO multiple objective

    @property
    def P(self) -> list[list[int]]:
        """Structure of the constraint matrix"""
        return [c.P for c in self.cons()]

    @property
    def Z(self) -> list[list[int]]:
        """Structure of the constraint matrix"""
        return [c.Z for c in self.cons()]

    @property
    def G(self) -> list[list[float]]:
        """Matrix of Variable coefficients for type:

        g < = 0
        """
        _G = [[0] * len(self.variables) for _ in range(len(self.leqcons()))]

        for n, c in enumerate(self.leqcons()):
            for x, a in zip(c.P, c.A):
                if x is not None:
                    _G[n][x] = a

        return _G

    @property
    def H(self) -> list[list[float]]:
        """Matrix of Variable coefficients for type:

        h = 0
        """
        _H = [[0] * len(self.variables) for _ in range(len(self.eqcons()))]
        for n, c in enumerate(self.eqcons()):
            for x, a in zip(c.P, c.A):
                if x is not None:
                    _H[n][x] = a
        return _H

    @property
    def NN(self) -> list[list[float]]:
        """Matrix of Variable coefficients for non negative cons"""
        _NN = [[0] * len(self.variables) for _ in range(len(self.variables))]

        for n, v in enumerate(self.variables):
            if v in self.nnvars():
                _NN[n][n] = -1
        return _NN

    @property
    def A_with_NN(self) -> list[list[float]]:
        """Matrix of Variable coefficients with non-negative constraints"""
        return self.A + self.NN

    @property
    def B_with_NN(self) -> list[float]:
        """RHS Parameter vector with non-negative constraints"""
        return self.B + [0] * len(self.nnvars())

    @property
    def CrA(self) -> list[list[float]]:
        """Critical Region A matrix"""
        CrA_UB = [[0] * len(self.thetas) for _ in range(len(self.thetas))]
        CrA_LB = [[0] * len(self.thetas) for _ in range(len(self.thetas))]

        for n in range(len(self.thetas)):
            CrA_UB[n][n] = 1.0
            CrA_LB[n][n] = -1.0

        CrA_ = []

        for n in range(len(self.thetas)):
            CrA_.append(CrA_UB[n])
            CrA_.append(CrA_LB[n])

        return CrA_

    @property
    def CrB(self) -> list[float]:
        """Critical Region RHS vector"""
        CrB_ = []
        for t in self.thetas:
            CrB_.append(t._[1])
            CrB_.append(-t._[0])

        return CrB_

    def make_A_df(self, longname: bool = False) -> DataFrame:
        """Create a DataFrame from the A matrix.

        Args:
            longname (bool, optional): Whether to use long names for variables. Defaults to False.

        Returns:
            DataFrame: Columns are the variables, rows are the constraints.
        """
        if longname:
            return DataFrame(
                self.A,
                columns=[v.longname for v in self.variables],
                index=[c.longname for c in self.cons()],
            )
        return DataFrame(
            self.A,
            columns=[v.name for v in self.variables],
            index=[c.name for c in self.cons()],
        )

    def make_B_df(self, longname: bool = False) -> DataFrame:
        """Create a DataFrame from the B vector.

        Args:
            longname (bool, optional): Whether to use long names for variables. Defaults to False.

        Returns:
            DataFrame: Has a single column, rows are the constraints.
        """
        if not has_pandas:
            raise ImportError("pip install gana[all] to get optional dependencies")
        if longname:
            index = [c.longname for c in self.cons()]
        else:
            index = [c.name for c in self.cons()]

        return DataFrame(self.B, columns=['RHS'], index=index)

    def make_C_df(self, longname: bool = False) -> DataFrame:
        """Create a DataFrame from the C matrix.

        Args:
            longname (bool, optional): Whether to use long names for variables. Defaults to False.

        Returns:
            DataFrame: Columns are the variables, Has a single row
        """

        if longname:
            columns = [v.longname for v in self.variables]
        else:
            columns = [v.name for v in self.variables]
        return DataFrame([self.C], columns=columns, index=['Minimize'])

    def make_df(self, longname: bool = False) -> DataFrame:
        """Create a DataFrame from the model.

        Args:
            longname (bool, optional): Whether to use long names for variables. Defaults to False.

        Returns:
            DataFrame: A DataFrame with the A matrix, B vector, and C vector.
        """

        if longname:
            index = ['Minimize'] + [c.longname for c in self.cons()]
            columns = [v.longname for v in self.variables] + ['RHS']
        else:
            index = ['Minimize'] + [c.name for c in self.cons()]
            columns = [v.name for v in self.variables] + ['RHS']
        data = []
        for n, d in enumerate(self.A):
            d.append(self.B[n])
            data.append(d)

        data = [self.C + [0]] + data

        return DataFrame(data, columns=columns, index=index)

    def make_CrA_df(self, longname: bool = False) -> DataFrame:
        """Creates a DataFrame from the Critical Region A matrix."""

        if longname:
            columns = [t.longname for t in self.thetas]
            index = sum([[t.longname] * 2 for t in self.thetas], [])
        else:
            columns = [t.name for t in self.thetas]
            index = sum([[t.name] * 2 for t in self.thetas], [])

        return DataFrame(self.CrA, columns=columns, index=index)

    def make_CrB_df(self, longname: bool = False) -> DataFrame:
        """Creates a DataFrame from the Critical Region RHS vector."""

        if longname:
            index = sum([[t.longname] * 2 for t in self.thetas], [])
        else:
            index = sum([[t.name] * 2 for t in self.thetas], [])

        return DataFrame(self.CrB, columns=['RHS'], index=index)

    def make_F_df(self, longname: bool = False) -> DataFrame:
        """Creates a DataFrame from the Theta coefficients matrix."""
        if longname:
            columns = [t.longname for t in self.thetas]
            index = [c.longname for c in self.cons()]
        else:
            columns = [t.name for t in self.thetas]
            index = [c.name for c in self.cons()]

        return DataFrame(self.F, columns=columns, index=index)

    # --------------------------------------------------
    #               Write
    # --------------------------------------------------

    def mps(self, name: str = None):
        """MPS File"""

        print(f'--- Generating {name or self.name}.mps')

        # 1 unit of whitespace
        ws = ' '
        # renumber the constraints based on order in .cons()
        # as opposed to order of declaration
        self.renumber()

        leq_cons = self.leqcons()
        eq_cons = self.eqcons()
        cont_vars = self.contvars()
        nn_vars = self.nnvars()
        bnr_vars = self.bnrvars()

        # _C = self.C
        # _A = self.A

        # write the MPS file
        with open(f'{name or self.name}.mps', 'w', encoding='utf-8') as f:

            # header: NAME          MODEL_NAME
            f.write(f'NAME{ws*10}{self.name.upper()}\n')

            # Here the constraint types are defined
            f.write('ROWS\n')

            if self.objectives:
                # the objective is: N   OBJECTIVE_NAME
                f.write(f'{ws}N{ws*3}{self.objectives[-1].mps()}\n')

            for c in leq_cons:
                # less than or equal constraints are: L   CONSTRAINT_NAME
                f.write(f'{ws}L{ws*3}{c.mps()}\n')

            for c in eq_cons:
                # equality constraints are: E   CONSTRAINT_NAME
                f.write(f'{ws}E{ws*3}{c.mps()}\n')

            # Here the variables are defined along with their coefficients
            # in each of the constraints that they feature in

            f.write('COLUMNS\n')
            for v in self.variables:
                # For each variable, we write:
                # V_NAME    CONSTRAINT_NAME    COEFFICIENT
                # for all variables, these are ordered
                # as they are added based on declaration
                vs = len(v.mps())
                # for constraints/functions/objectives that they feature in
                for c in v.cons_by:
                    # this captures the length of the variable name
                    # variable names are just Vn where n is order of precedence
                    vfs = len(c.mps())
                    f.write(ws * 4)
                    f.write(v.mps())
                    f.write(ws * (10 - vs))
                    f.write(c.mps())
                    f.write(ws * (10 - vfs))

                    # C variable coefficients are a vector
                    # if isinstance(c, O):
                    #     f.write(f'{c.function[0].matrix[v.n]}')
                    # else:
                    #     f.write(f'{c.matrix[v.n]}')
                    f.write(f'{c.matrix[v.n]}')

                    # f.write(f'{c.A[c.P.index(v.n)]}')

                    f.write('\n')

                for o in v.min_by:
                    # this captures the length of the variable name
                    # variable names are just Vn where n is order of precedence
                    vfs = len(o.mps())
                    f.write(ws * 4)
                    f.write(v.mps())
                    f.write(ws * (10 - vs))
                    f.write(o.mps())
                    f.write(ws * (10 - vfs))

                    f.write(f'{o.function[0].matrix[v.n]}')

                    # f.write(f'{c.A[c.P.index(v.n)]}')
                    f.write('\n')

            # This gives the right-hand side of the constraints
            f.write('RHS\n')
            for n, c in enumerate(leq_cons + eq_cons):
                # For each constraint, we write:
                # RHSn    CONSTRAINT_NAME    RHS_VALUE
                f.write(ws * 4)
                f.write(f'RHS{n}')
                f.write(ws * (10 - len(f'RHS{n+1}')))
                f.write(c.mps())
                f.write(ws * (10 - len(c.mps())))
                f.write(f'{c.B}')
                f.write('\n')

            f.write('BOUNDS\n')
            # for continuous variables that are nonnegative, we write:
            # LO BND1    VARIABLE_NAME    0
            for v in nn_vars:
                if v in cont_vars:
                    f.write(f'{ws}LO{ws}BND1{ws*4}{v.mps()}{ws*8}{0}\n')

            # for integer variables that are binary, we write:
            # BV BND1    VARIABLE_NAME
            for v in bnr_vars:
                f.write(f'{ws}BV{ws}BND1{ws*5}{v.mps()}\n')

            # CLOSE the MPS file
            f.write('ENDATA')

    def lp(self):
        """LP File"""
        m = self.gurobi()
        m.write(f'{self.name}.lp')

    # --------------------------------------------------
    #               Optimize
    # --------------------------------------------------
    def opt(self, using: str = 'gurobi'):
        """Solve the program"""

        if using == 'gurobi':
            m = self.gurobi()

            print(f'--- Optimizing {self} using {using}')
            m.optimize()
            try:
                print('--- Solution found. Use .sol() to display it')

                self.X[self.n_sol] = [v.X for v in m.getVars()]

                for v, val in zip(self.variables, self.X[self.n_sol]):

                    v.X[self.n_sol] = val

                for c in self.constraint_sets:
                    c.function.eval(n_sol=self.n_sol)

                self.objectives[-1].X = m.ObjVal
                self.optimized = True

                print('--- Creating Solution object, check.solution')

                self.solution[self.n_sol] = self.birth_solution()

                self.n_sol += 1

            except AttributeError:
                print('!!! No solution found. Check the model.')

    def lb(self, function: V | Func):
        """Finds the lower bound of a variable or function"""
        # set the objective to minimizing the variable
        setattr(self, f'min({function})', inf(function))
        self.opt()

    def ub(self, function: V | Func):
        """Finds the upper bound of a variable or function"""
        # set the objective to maximizing the variable
        setattr(self, f'max({function})', sup(function))
        self.opt()

    def obj(self):
        """Objective Values"""
        if len(self.objectives) == 1:
            return self.objectives[0].X
        return {o: o.X for o in self.objectives}

    # def slack(self):
    #     """Slack in each constraint"""
    #     return {c: c._ for c in self.leqcons()}

    def sol(self, n_sol: int = 0, slack: bool = True, compare=False):
        """Print sol"""

        if not self.optimized:
            return r'Use .opt() to generate solution'

        display(Markdown(rf'# Solution for {self.name}'))

        print()
        display(Markdown(r'## Objective'))

        self.objectives[n_sol].sol()

        print()
        display(Markdown(r'## Variables'))

        for v in self.variable_sets:
            v.sol(n_sol=n_sol, compare=compare)

        if slack:
            print()
            display(Markdown(r'## Constraint Slack'))
            for c in self.leqcons():
                c.sol(n_sol=n_sol, compare=compare)

    def birth_solution(self):
        """Makes a solution object for the program"""

        solution = Solution(self.name + '_solution_' + str(self.n_sol))
        solution.update(self.variables, n_sol=self.n_sol)

        return solution

    # # Displaying the program
    # def latex(self, descriptive: bool = False):
    #     """Display LaTeX"""

    #     for s in self.sets.index:
    #         display(s.latex(True))

    #     for o in self.objectives:
    #         display(o.latex())

    #     if descriptive:
    #         for c in self.cons():
    #             display(c.latex())

    #     else:
    #         for c in self.sets.cons():
    #             display(c.latex())

    #         for c in self.cons():
    #             if not c.parent:
    #                 display(c.latex())

    def show(
        self,
        descriptive: bool = False,
        nncons: bool = False,
        categorical: bool = False,
        category: str = None,
    ):
        """Pretty Print"""

        display(Markdown(rf'# Mathematical Program for {self}'))

        if category:
            categorical = True

        if self.index_sets:
            print()
            display(Markdown(r'## Index Sets'))

            for i in self.index_sets:
                if len(i) != 0 and i.case != ICase.SELF:
                    i.show(True)

        if self.objectives:
            print()
            display(Markdown(r'## Objective'))

            for o in self.objectives:
                o.show()

        if descriptive:

            print()
            display(Markdown(r'## s.t.'))

            if categorical:
                # gather the categories if not already done
                categories: dict[str, list[C]] = {}
                fcategories: dict[str, list[Func]] = {}
                for c in self.cons():
                    if c.category not in categories:
                        categories[c.category] = []
                    categories[c.category].append(c)
                if category and category in categories:
                    sorted_categories = [category]
                else:
                    sorted_categories = sorted(categories.keys())
                self.categories = categories

                for f in self.function_sets:
                    if f.category not in fcategories:
                        fcategories[f.category] = []
                    fcategories[f.category].append(f)
                if category and category in fcategories:
                    sorted_fcategories = [category]
                else:
                    sorted_fcategories = sorted(fcategories.keys())
                self.fcategories = fcategories

                for category in sorted_categories:
                    display(Markdown(rf'### {category} Constraints'))
                    for c in categories[category]:
                        c.show()
                for category in sorted_fcategories:
                    display(Markdown(rf'### {category} Functions'))
                    for f in fcategories[category]:
                        f.show()

            else:

                if self.leqcons():
                    display(Markdown(r'### Inequality Constraints'))
                    for c in self.leqcons():
                        c.show()
                if self.eqcons():
                    display(Markdown(r'### Equality Constraints'))
                    for c in self.eqcons():
                        c.show()

                if self.nncons() and nncons:
                    display(Markdown(r'### Non-Negative Constraints'))
                    for c in self.nncons():
                        c.show()

                if self.functions:
                    print()
                    display(Markdown(r'## Functions'))
                    for f in self.functions:
                        f.show()

        else:

            # if self.sets.nncons():
            #     print()
            #     display(Markdown(r'## Non-Negative Variables'))
            #     self.sets.I_nn.show()

            print()
            display(Markdown(r'## s.t.'))

            if categorical:
                # gather the categories if not already done
                categories_sets: dict[str, list[C]] = {}
                fcategories_sets: dict[str, list[Func]] = {}

                for c in self.constraint_sets:
                    if c.category not in categories_sets:
                        categories_sets[c.category] = []
                    categories_sets[c.category].append(c)
                if category and category in categories_sets:
                    sorted_categories = [category]
                else:
                    sorted_categories = sorted(categories_sets.keys())
                self.categories_sets = categories_sets

                for f in self.function_sets:
                    if f.category not in fcategories_sets:
                        fcategories_sets[f.category] = []
                    fcategories_sets[f.category].append(f)
                if category and category in fcategories_sets:
                    sorted_fcategories = [category]
                else:
                    sorted_fcategories = sorted(fcategories_sets.keys())
                self.fcategories_sets = fcategories_sets

                for category in sorted_categories:
                    display(Markdown(rf'### {category} Constraint Sets'))
                    for c in categories_sets[category]:
                        c.show()

                for category in sorted_fcategories:
                    display(Markdown(rf'### {category} Function Sets'))
                    for f in fcategories_sets[category]:
                        f.show()

            else:

                if self.leqcons_sets:
                    print()
                    display(Markdown(r'### Inequality Constraint Sets'))
                    for c in self.leqcons_sets:
                        c.show()
                if self.eqcons_sets:
                    print()
                    display(Markdown(r'### Equality Constraint Sets'))
                    for c in self.eqcons_sets:
                        c.show()

                if self.function_sets:
                    print()
                    display(Markdown(r'## Functions'))
                    for f in self.function_sets:
                        f.show()

    def draw(self, variable: V, n_sol: int = 0):
        """Plots the solution for a variable"""
        self.solution[n_sol].draw(variable)

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    def __str__(self):
        return rf'{self.name}'

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(str(self))

    # def __add__(self, other: Self):
    #     """Add two programs"""

    #     if not isinstance(other, Prg):
    #         raise ValueError('Can only add programs')

    #     prg = Prg(name=rf'{self.name}')

    #     for i in (
    #         self.sets.index
    #         + other.sets.index
    #         + self.sets.variable
    #         + other.sets.variable
    #         + self.sets.parameter
    #         + other.sets.parameter
    #     ):
    #         if not i.name in prg.names:
    #             setattr(prg, i.name, i)
    #         else:
    #             if isinstance(i, I) and i.mutable:
    #                 setattr(prg, i.name, getattr(prg, i.name) | i)

    #     for i in (
    #         self.sets.function
    #         + other.sets.function
    #         + self.sets.leqcons()
    #         + self.sets.eqcons()
    #         + other.sets.leqcons()
    #         + other.sets.eqcons()
    #         + self.objectives
    #         + other.objectives
    #     ):
    #         if not i.name in prg.names:
    #             setattr(prg, i.pname, i)

    #     return prg
    # -----------------------------------------------------
    #                    Export
    # -----------------------------------------------------

    def ppopt(self) -> MPLP_Program:
        """Convert the program to a ppopt.MPLP_Program"""

        # A is the matrix of variable coefficients (including nn constraints)
        # b is the RHS vector (including nn constraints)
        # c is the objective coefficients
        # A_t is the critical region A matrix
        # b_t is the critical region RHS vector
        # F is the matrix of theta coefficients (including nn constraints)
        # H are the parameteric objective coefficients

        return MPLP_Program(
            A=nparray(self.A + self.NN),
            b=nparray([[i] for i in self.B] + [[0]] * self.n_variables),
            c=nparray([[i] for i in self.C]),
            A_t=nparray(self.CrA),
            b_t=nparray([[i] for i in self.CrB]),
            F=nparray(self.F + [[0] * self.n_thetas] * self.n_variables),
            H=npzeros((self.n_variables, self.n_thetas)),
            equality_indices=[c.n for c in self.cons() if c.eq],
        )

    # def pyomo(self):
    #     """Pyomo Model"""
    #     if has_pyomo:
    #         m = PyoModel()

    #         for index_set in self.index_sets:
    #             setattr(m, index_set.name, index_set.pyomo())

    #         for v in self.variable_sets:
    #             setattr(m, v.name, v.pyomo())

    #         # for c in self.constraint_sets:

    #         # for c in self.conssets:
    #         #     setattr(m, c.name, c.pyomo(m))

    #         return m
    #     print(
    #         'pyomo is an optional dependency, pip install gana[all] to get optional dependencies'
    #     )

    def gurobi(self):
        """Gurobi Model"""

        self.mps()
        print(f'--- Creating gurobi model for {self}')
        return gpread(f'{self}.mps')



"""Paramter Set"""

from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING, Self
from warnings import warn

from IPython.display import Math, display

from .birth import make_T
from .cases import Elem, PCase
from .function import F
from .index import I
from .variable import V

if TYPE_CHECKING:
    from .theta import T

try:
    from pyomo.environ import Param as PyoParam

    has_pyomo = True
except ImportError:
    has_pyomo = False

try:
    from sympy import Idx, IndexedBase, symbols

    has_sympy = True
except ImportError:
    has_sympy = False


class P:
    """
    Ordered set of parameters.

    Does not support `inf` or `nan` values.

    :param index: Indices of the parameter set.
    :type index: tuple[I], optional
    :param _: List of parameters. All values are converted to float.
    :type _: list[int | float], optional
    :param mutable: If the parameter set is mutable.
    :type mutable: bool, optional
    :param tag: Tag/details
    :type tag: str, optional

    :ivar index: Index of the parameter set
    :vartype index: I
    :ivar _: List of parameters (converted to float)
    :vartype _: list[int | float]
    :ivar mutable: If the parameter set is mutable
    :vartype mutable: bool
    :ivar tag: Tag/details
    :vartype tag: str
    :ivar name: Name, set by the program
    :vartype name: str
    :ivar n: Number id, set by the program
    :vartype n: int
    :ivar map: Index to parameter mapping
    :vartype map: dict[X | Idx, Var]
    :ivar case: Special case of the parameter set
    :vartype case: PCase

    :raises ValueError: If `!=` operator is used with any type other than `P`
    :raises ValueError: If the parameter values and the length of indices do not match
    """

    # TODO update Errors in docstring
    # TODO add examples

    def __init__(
        self,
        *index: I,
        _: list[float] | float = None,
        mutable: bool = False,
        tag: str = None,
        ltx: str = None,
    ):
        # given at declaration
        self.tag = tag
        self._ltx = ltx
        self.mutable = mutable

        # name will be set by the program later
        # if dummy index, the name is set to 'φ' (phi)
        self.name = ''

        # special case of the parameter
        self.case: PCase = PCase.SET
        # set the index
        # self.index: tuple[I] | set[tuple[I]] = index

        if any([isinstance(i, tuple) for i in index]):
            self.n_splices = len(index)
            # if index is a set of indices,
            # needs to be done for each index
            _index = []
            _map = {}
            for idx in index:
                _index.append(tuple([i if not isinstance(i, V) else [i] for i in idx]))

            # iterates over each individual index
            # and creates a mapping for it
            for idx in _index:
                for i in product(*idx):
                    _map[i] = None
            _index = set(_index)

        else:
            # number of splices of the index set
            self.n_splices = 1
            # if not set
            _index = tuple([i if not isinstance(i, V) else [i] for i in index])

            if _index:
                _map = {i: None for i in product(*_index)}

            else:
                _map = {}

        self.index: tuple[I] | set[tuple[I]] = _index
        self.map: dict[I, V] = _map

        # contains the set of parameters
        if _ is None:
            self._ = []
        else:
            self._: list[float | int] = _  # always a list of parameters

        # set by the program
        # this is the nth parameter declared
        self.n: int = None

        # this helps in the index check when calling functions
        self.elements = [self]

        if isinstance(self._, (int, float)):
            # if int or float is passed it is a single number
            # names for these are generated automatically
            # in __str__()
            self.name = str(_)  # set the name to the number
            if _ > 0:
                # postive number
                self.case = PCase.NUM
            elif _ < 0:
                # negative number
                self.case = PCase.NEGNUM
            else:
                # 0
                self.case = PCase.ZERO

            if self.index:
                # if index is passed
                # make length equal to index
                self._ = [float(_)] * len(self.map)
            else:
                self.index = (I(size=1, dummy=True),)
                self._ = [float(_)]

        else:
            # if some sort of iterable is passed
            # preferably a list
            # set by program
            if not self.index and self._:
                # if index is not passed
                # make a dummy index
                self.index = (I(size=len(self._), dummy=True),)
                # here the map needs to be remade
                self.map = {i: None for i in list(product(*self.index))}
                self.name = 'φ'  # set the name to phi
            self._ = [float(p) for p in self._]

        # fill in the values
        for n, k in enumerate(self.map):
            self.map[k] = self._[n]

    @property
    def args(self) -> dict[str, str | bool]:
        """Return the arguments of the parameter set"""
        return {'tag': self.tag, 'ltx': self.ltx, 'mutable': self.mutable}

    @property
    def ltx(self) -> str:
        """LaTeX representation of the variable set"""
        if self._ltx:
            return r'{' + self._ltx + r'}'
        # if user has not set the LaTeX representation
        # the name becomes the latex representation
        if self.name:
            return r'{' + self.name.replace('_', r'\_') + r'}'
        return self._ltx

    # -----------------------------------------------------
    #                   Matrix
    # -----------------------------------------------------
    @property
    def A(self) -> list[list[float]]:
        """Generate a diagonal matrix representation of the variable set"""
        return [[self._[i]] for i in range(len(self))]

    # -----------------------------------------------------
    #                    Printing
    # -----------------------------------------------------

    def latex(self) -> str:
        """LaTeX representation"""

        if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
            return str(self)

        index = (
            r'_{'
            + rf'{self.index}'.replace('(', '')
            .replace(')', '')
            .replace('[', '{')
            .replace(']', '}')
            + r'}'
        )

        if len(self.index) == 1:
            # if there is a single index element
            # then a comma will show up in the end, replace that
            return self.ltx + index.replace(',', '')  # type: ignore

        return self.ltx + index

    def show(self, descriptive: bool = False):
        """Display the variables
        Args:
            descriptive (bool, optional): If True, shows all parameters. Defaults to False.
        """
        if descriptive:
            # just print out the parameters
            for p in self._:
                print(p)

        display(Math(self.latex()))

    # -----------------------------------------------------
    #                    Value
    # -----------------------------------------------------

    def __neg__(self):
        if self.case == PCase.ZERO:
            # if zero return self
            return self

        if self.case in [PCase.NUM, PCase.NEGNUM]:
            # if number return a negation

            return P(*self.index, _=-self._[-1])

        # else negate the number and return a new parameter set
        p = P(*self.index, _=[-i for i in self._], **self.args)
        if self.case == PCase.NEGSET:
            # if this is already a negated set
            # make it a normal set now
            p.name = f'{self.name[1:]}'  # remove the negative sign
            p.case = PCase.SET
        else:
            # if this is a normal set
            # make it a negated set
            p.case = PCase.NEGSET
            p.name = f'-{self}'  # add the negative sign
        # return the new parameter set
        return p

    def __pos__(self):

        if self.case == PCase.NEGNUM:
            # if it is a negative number
            # return a positive number
            return P(*self.index, _=-self._[-1], **self.args)

        if self.case == PCase.NEGSET:
            p = P(*self.index, _=[-i for i in self._], **self.args)
            p.name = self.name[1:]  # remove the negative sign
            return p

        # if self.case in [PCase.ZERO, PCase.SET, PCase.NUM]:
        # if it is zero, a normal set, or a number
        # return itself
        return self

    def __abs__(self):
        if self.case == PCase.ZERO:
            # if zero return self
            return self

        if self.case == PCase.SET:
            # if it is a set return a normal set
            p = P(*self.index, _=[abs(i) for i in self._], **self.args)
            p.name = f'|{self.name}|'
            return p

        if self.case in [PCase.NEGNUM, PCase.NEGSET]:
            # if it is a negative number or a negated set
            # return a positive number or a normal set
            return -self
        # else just itself
        return self

    def __float__(self):
        if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
            return self._[-1]
        return self

    # -----------------------------------------------------
    #                    Operators
    # -----------------------------------------------------
    # --- Handling basic operations----
    # if there is a zero on the left, just return P
    # if the other is a parameter, add the values
    # if the other is a function/variable, return a function

    # r<operation>
    # for the right hand side operations
    # they only kick in when the left hand side operator
    # does not have the operation/the operation did not work
    # in this case, we just do the equivalent self

    def __add__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:
        if other is None:
            # if adding with None
            return self

        if self.case == PCase.ZERO:
            # if self is zero
            # return other
            return other

        if isinstance(other, (int, float)):
            # if adding to a number
            if other in [0, 0.0]:
                # if adding zero, return self
                return self

            if self.case in [PCase.NEGNUM, PCase.NUM]:
                # P (number) - other
                # the number will be set to NEGNUM, NUM
                if self._[-1] + other in [0, 0.0]:
                    # if number returns zero, return int 0
                    return 0

                return P(*self.index, _=self._[-1] + other, **self.args)

            # if set or negative set
            # just add the number to each value
            # and return new parameter set
            p = P(*self.index, _=[i + other for i in self._], **self.args)
            p.name = f'({self}+{other})'
            return p

        if isinstance(other, list):
            # lengths should match
            if self.case in [PCase.NUM, PCase.NEGNUM]:
                p = P(*self.index, _=[self._[-1] + i for i in other], **self.args)

            else:
                if len(self) != len(other):
                    warn(
                        f"Index mismatch {self} + {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                    )
                # if list, just zip through the values and add
                p = P(
                    *self.index, _=[i + j for i, j in zip(self._, other)], **self.args
                )
            # change the name to add that something from a list has been added
            p.name = f'({self}+φ)'
            return p

        if isinstance(other, tuple):
            # if tuple, make T
            # and make a function
            return F(
                one=make_T(other, index=self.index),
                add=True,
                two=self,
                one_type=Elem.T,
                two_type=Elem.P,
            )

        if isinstance(other, P):
            # lengths should match

            if other.case == PCase.ZERO:
                # if other is zero
                return self

            if self.case in [PCase.NUM, PCase.NEGNUM]:

                if other.case in [PCase.NUM, PCase.NEGNUM]:
                    # if both are numbers
                    # let P handle the case
                    if self._[-1] + other._[-1] in [0, 0.0]:
                        # if number returns zero, return int 0
                        return 0

                    return P(*self.index, _=self._[-1] + other._[-1], **self.args)
                # else make a new parameter set

                p = P(*self.index, _=[self._[-1] + i for i in other._], **self.args)
                p.name = f'({self}+{other})'
                return p

            if other.case in [PCase.NUM, PCase.NEGNUM]:
                # if other is a number, add other to every value in the set
                p = P(*self.index, _=[i + other._[-1] for i in self._], **self.args)
                p.name = f'({self}+{other})'
                return p

            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} + {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            # if either one of them is general parameter
            p = P(*self.index, _=[i + j for i, j in zip(self._, other._)], **self.args)
            p.name = f'({self}+{other})'
            return p

        # if not parameter, let the other handle elements operator handle the addition
        if self.case in [PCase.NEGNUM, PCase.NEGSET]:
            # if negative number, handle using sub, but keep parameter at two
            # (-P) + E = E - P
            return F(one=other, sub=True, two=-self, two_type=Elem.P)

        # keep additive or subtractive parameter at two
        # P + E = E + P
        return F(one=other, add=True, two=self, two_type=Elem.P)

    def __radd__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:
        # let __add__() handle the addition
        return self + other

    def __sub__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P | F:
        if other is None:
            # if adding with None
            return self

        if self.case == PCase.ZERO:
            # if self is zero
            # return negative of other
            return -other

        if isinstance(other, (int, float)):
            # if adding to a number
            if other in [0, 0.0]:
                # if adding zero, return self
                return self

            if self.case in [PCase.NEGNUM, PCase.NUM]:

                if self._[-1] - other == 0:
                    # if number returns zero, return int 0
                    return 0

                # let P make a .NUM type
                return P(*self.index, _=self._[-1] - other, **self.args)

            p = P(*self.index, _=[i - other for i in self._], **self.args)
            p.name = f'({self}-{other})'
            return p

        if isinstance(other, list):
            # if list, just zip through the values and add

            if self.case in [PCase.NUM, PCase.NEGNUM]:
                p = P(*self.index, _=[self._[-1] - i for i in other], **self.args)
            else:
                if len(self) != len(other):
                    warn(
                        f"Index mismatch {self} - {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                    )
                p = P(
                    *self.index, _=[i - j for i, j in zip(self._, other)], **self.args
                )
                # change the name to add that something from a list has been added
            p.name = f'({self}-φ)'
            return p

        if isinstance(other, tuple):
            # if tuple, make T
            # and make a function
            return F(
                one=make_T(other, index=self.index),
                sub=True,
                two=self,
                one_type=Elem.T,
                two_type=Elem.P,
            )

        if isinstance(other, P):

            if other.case == PCase.ZERO:
                # if other is zero
                return self

            if self.case in [PCase.NUM, PCase.NEGNUM]:

                if other.case in [PCase.NUM, PCase.NEGNUM]:
                    # if both are numbers
                    # let P handle the case
                    if self._[-1] - other._[-1] in [0, 0.0]:
                        # if number returns zero, return int 0
                        return 0
                    # let P handle the case based on the value that is determines
                    return P(*self.index, _=self._[-1] - other._[-1], **self.args)

                p = P(*self.index, _=[self._[-1] - i for i in other._], **self.args)
                p.name = f'({self}-{other.name})'
                return p

            if other.case in [PCase.NUM, PCase.NEGNUM]:
                # if other is a number subtract every value in self with the number
                p = P(*self.index, _=[i - other._[-1] for i in self._], **self.args)
                p.name = f'({self}-{other})'
                return p

            if len(self) != len(other):
                raise ValueError(
                    f"Index mismatch {self} - {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )

            # if both are just general parameters
            # zip through the values and subtract
            p = P(*self.index, _=[i - j for i, j in zip(self._, other._)], **self.args)
            p.name = f'({self}-{other})'
            return p

        # if not parameter, let the other handle elements operator handle the addition
        if self.case in [PCase.NEGNUM, PCase.NEGSET]:
            # if negative number, handle using function sub, but keep parameter at two
            # -P - E = -E - P
            return F(one=-other, sub=True, two=self, two_type=Elem.P)

        # keep additive or subtractive parameter at two
        # P - E = -E + P
        return F(one=-other, add=True, two=self, two_type=Elem.P)

    def __rsub__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P | F:
        # let negation and __add__ handle the subtraction
        return -self + other

    def __mul__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:

        if other is None:
            # if multiplying by nothing, return nothing
            return None

        if self.case == PCase.ZERO:
            # if self is zero, return zero
            return 0

        if isinstance(other, (int, float)):
            # if multiplying with a number
            if other in [1, 1.0]:
                # by unity, return itself
                return self
            if other in [0, 0.0]:
                # by zero, return 0
                return 0.0

            if self.case in [PCase.NEGNUM, PCase.NUM]:
                # if self is a number, just find the product
                # let P handle the rest
                return P(*self.index, _=self._[-1] * other, **self.args)

            # else multiply the number to each value
            p = P(*self.index, _=[i * other for i in self._], **self.args)
            p.name = f'{self}*{other}'
            return p

        if isinstance(other, list):
            # if list, just zip through the values and multiply
            if self.case in [PCase.NUM, PCase.NEGNUM]:
                p = P(*self.index, _=[self._[-1] * i for i in other], **self.args)
            else:
                if len(self) != len(other):
                    warn(
                        f"Index mismatch {self} * {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                    )
                p = P(
                    *self.index, _=[i * j for i, j in zip(self._, other)], **self.args
                )
            p.name = f'{self}*φ'
            return p

        if isinstance(other, tuple):
            # if tuple, make T
            # return a scaled T
            if self.case == PCase.ZERO:
                # return int zero
                return 0

            if self.case in [PCase.NUM, PCase.NEGNUM]:
                # if self is a number, scale the tuple by the number
                # and make a T

                t = make_T(
                    tuple([other[0] * self._[-1], other[1] * self._[-1]]), self.index
                )
                t.name = f'{self}*θ'
                return t

            # otherwise self is a set
            # scale the tuple by each value in the set
            t = make_T(
                [tuple([i * other[0], i * other[1]]) for i in self._], self.index
            )
            t.name = f'{self}*θ'
            return t

        if isinstance(other, P):

            if other.case == PCase.ZERO:
                # if self or other is zero, return int zero
                return 0

            if self.case in [PCase.NUM, PCase.NEGNUM]:
                # if self is a number, just find the product
                # let P handle the rest
                if other.case in [PCase.NUM, PCase.NEGNUM]:

                    if other._[0] in [0, 0.0]:
                        # if other is zero, return int zero
                        return 0

                    return P(*self.index, _=self._[-1] * other._[-1], **self.args)

                p = P(*self.index, _=[self._[-1] * i for i in other._], **self.args)
                p.name = f'{self}*{other}'
                return p

            if other.case in [PCase.NUM, PCase.NEGNUM]:
                # if other is a number, find the product of every value in self with the number
                p = P(*self.index, _=[i * other._[-1] for i in self._], **self.args)
                p.name = f'{self}*{other}'
                return p

            if len(self) != len(other):
                raise ValueError(
                    f"Index mismatch {self} * {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            # if both are just general parameters
            # else zip through the values and multiply
            p = P(*self.index, _=[i * j for i, j in zip(self._, other._)], **self.args)
            p.name = rf"{self}*{other}"
            return p

        return other * self
        # TODO - handle multiplication with other types
        # call multiplication function
        # if isinstance(other, F):
        #     # if other is a function, use it
        #     # let F handle the multiplication
        #     if other.add:
        #         # if the function is an addition, return a function
        #         if self.case in [PCase.NEGNUM, PCase.NEGSET]:
        #             if other.two_type == Elem.P:
        #                 # if the other is a parameter, make it a subtraction
        #                 return F(one=self * other.one, sub=True, two=self * other.two)
        #             # make it a subtraction
        #             return F(one=self * other.one, sub=True, two=self * other.two)

        #         return F(one=self * other.one, add=True, two=self * other.two)

        # return F(one=self, mul=True, two=other, one_type=Elem.P)

    def __rmul__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:
        # multiplication is commutative
        return self * other

    def __truediv__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:
        try:
            if other is None:
                # if dividing by nothing, return nothing
                raise ValueError(
                    "Cannot divide by None. Please provide a valid parameter, variable, or number."
                )

            if self.case == PCase.ZERO:
                # if self is zero, return zero
                # take every opportunity to return int 0
                return 0

            if isinstance(other, (int, float)):
                # if dividing by a number
                if other in [0, 0.0]:
                    # if dividing by zero, raise an error
                    raise ZeroDivisionError(f'{self} cannot be divided by zero.')

                if other in [1, 1.0]:
                    return self

                if self.case in [PCase.NEGNUM, PCase.NUM]:
                    # if self is a number, just find the division
                    # let P handle the rest
                    return P(*self.index, _=self._[-1] / other, **self.args)

                p = P(*self.index, _=[i / other for i in self._], **self.args)
                p.name = f'{self}/{other}'
                return p

            if isinstance(other, list):
                if isinstance(other[0], tuple):
                    raise ValueError(
                        "Cannot divide by a list of tuples. Please provide a valid parameter, variable, or number."
                    )

                if self.case in [PCase.NUM, PCase.NEGNUM]:
                    # if self is a number, divide the number by each value in the list
                    p = P(*self.index, _=[self._[-1] / i for i in other], **self.args)

                else:
                    if len(self) != len(other):
                        warn(
                            f"Index mismatch {self} / {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                        )
                    # if list, just zip through the values and divide
                    p = P(
                        *self.index,
                        _=[i / j for i, j in zip(self._, other)],
                        **self.args,
                    )
                    # change the name to add that something from a list has been added
                    p.name = f'{self}/φ'
                    return p

            if isinstance(other, tuple):
                # dividing by parametric variable
                raise ValueError(
                    f"Cannot divide {self} by a tuple. Please provide a valid parameter, variable, or number."
                )

            if isinstance(other, P):

                if other.case == PCase.ZERO:
                    raise ZeroDivisionError(f'{self} cannot be divided by zero.')

                if self.case in [PCase.NEGNUM, PCase.NUM]:
                    if other.case in [PCase.NUM, PCase.NEGNUM]:
                        # if both are numbers
                        # let P handle the case
                        return P(*self.index, _=self._[-1] / other._[-1], **self.args)

                    p = P(*self.index, _=[self._[-1] / i for i in other._], **self.args)
                    p.name = f'{self}/{other}'
                    return p

                if other.case in [PCase.NUM, PCase.NEGNUM]:
                    # if other is a number, divide every value in self by the number
                    p = P(*self.index, _=[i / other._[-1] for i in self._], **self.args)
                    p.name = f'{self}/{other}'
                    return p

                if len(self) != len(other):
                    raise ValueError(
                        f"Index mismatch {self} / {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                    )
                # if both are just general parameters
                # zip through the values and divide
                p = P(*self.index, _=[i / j for i, j in zip(self._, other._)])
                p.name = f'{self}/{other}'
                return p

            # P / E: not handled yet
            raise ValueError(
                f"Cannot divide {self} by {other}. Nonlinear operations are not supported for {type(other).__name__} objects."
            )

        except ZeroDivisionError as e:
            # handle division by zero
            raise ZeroDivisionError(f'{self} cannot be divided by zero.') from e

    def __rtruediv__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:
        # just do this operation instead
        return (1 / self) * other

    def __floordiv__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:
        if other is None:
            # if dividing by nothing, return nothing
            raise ValueError(
                "Cannot divide by None. Please provide a valid parameter, variable, or number."
            )

        if self.case == PCase.ZERO:
            # if self is zero, return zero
            return 0

        if isinstance(other, (int, float)):
            # if dividing by a number
            if other in [0, 0.0]:
                # if dividing by zero, raise an error
                raise ZeroDivisionError(f'{self} cannot be divided by zero.')

            if other in [1, 1.0]:
                return self

            if self.case in [PCase.NEGNUM, PCase.NUM]:
                # if self is a number, just find the division
                # let P handle the rest
                return P(*self.index, _=self._[-1] // other, **self.args)

            p = P(*self.index, _=[i // other for i in self._], **self.args)
            p.name = f'{self}//{other}'
            return p

        if isinstance(other, list):
            if isinstance(other[0], tuple):
                raise ValueError(
                    "Cannot divide by a list of tuples. Please provide a valid parameter, variable, or number."
                )

            if self.case in [PCase.NUM, PCase.NEGNUM]:
                p = P(*self.index, _=[self._[-1] // i for i in other], **self.args)
            else:
                if len(self) != len(other):
                    warn(
                        f"Index mismatch {self} // {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                    )
                # if list, just zip through the values and divide
                p = P(
                    *self.index, _=[i // j for i, j in zip(self._, other)], **self.args
                )
            # change the name to add that something from a list has been added
            p.name = f'{self}//φ'
            return p

        if isinstance(other, tuple):
            # dividing by parametric variable
            raise ValueError(
                f"Cannot divide {self} by a tuple. Please provide a valid parameter, variable, or number."
            )

        if isinstance(other, P):

            if other.case == PCase.ZERO:
                raise ZeroDivisionError(f'{self} cannot be divided by zero.')

            if self.case in [PCase.NEGNUM, PCase.NUM]:
                if other.case in [PCase.NUM, PCase.NEGNUM]:
                    # if both are numbers
                    # let P handle the case
                    return P(*self.index, _=self._[-1] // other._[-1], **self.args)

                p = P(*self.index, _=[self._[-1] // i for i in other._], **self.args)
                p.name = f'{self}//{other}'
                return p
            if other.case in [PCase.NUM, PCase.NEGNUM]:
                # if other is a number, divide every value in self by the number
                p = P(*self.index, _=[i // other._[-1] for i in self._], **self.args)
                p.name = f'{self}//{other}'
                return p
            if len(self) != len(other):
                raise ValueError(
                    f"Index mismatch {self} // {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            # if both are just general parameters
            # zip through the values and divide
            p = P(*self.index, _=[i // j for i, j in zip(self._, other._)], **self.args)
            p.name = f'{self}//{other}'
            return p

        # else let the other handle the division
        return self // other

    def __mod__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:
        if other is None:
            # if dividing by nothing, return nothing
            raise ValueError(
                "Cannot divide by None. Please provide a valid parameter, variable, or number."
            )
        if self.case == PCase.ZERO:
            # if self is zero, return zero
            return 0

        if isinstance(other, (int, float)):
            # if dividing by a number
            if other in [0, 0.0]:
                # if dividing by zero, raise an error
                raise ZeroDivisionError(f'{self} cannot be divided by zero.')

            if other in [1, 1.0]:
                # if dividing by 1, return self
                return self

            if self.case in [PCase.NEGNUM, PCase.NUM]:
                # if self is a number, just find the modulus
                # let P handle the rest
                p = P(*self.index, _=self._[-1] % other, **self.args)
                p.name = f'{self} % {other}'
                return p

            p = P(*self.index, _=[i % other for i in self._], **self.args)
            p.name = f'{self} % {other}'
            return p

        if isinstance(other, list):
            if isinstance(other[0], tuple):
                raise ValueError(
                    "Cannot divide by a list of tuples. Please provide a valid parameter, variable, or number."
                )

            if self.case in [PCase.NUM, PCase.NEGNUM]:
                # if self is a number, find the mod of self and each number in the list
                p = P(*self.index, _=[self._[-1] % i for i in other], **self.args)

            else:
                if len(self) != len(other):
                    warn(
                        f"Index mismatch {self} % {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                    )
                # if list and self is a set, just zip through
                p = P(
                    *self.index, _=[i % j for i, j in zip(self._, other)], **self.args
                )
                # change the name to add that something from a list has been added
            p.name = f'{self} % φ'
            return p

        if isinstance(other, tuple):
            # dividing by parametric variable
            raise ValueError(
                f"Cannot divide {self} by a tuple. Please provide a valid parameter, variable, or number."
            )

        if isinstance(other, P):

            if self.case == PCase.ZERO:
                # if self is zero, return zero
                return 0

            if other.case == PCase.ZERO:
                raise ZeroDivisionError(f'{self} cannot be divided by zero.')

            if self.case in [PCase.NEGNUM, PCase.NUM]:
                if other.case in [PCase.NUM, PCase.NEGNUM]:
                    # if both are numbers
                    # let P handle the case
                    return P(*self.index, _=self._[-1] % other._[-1], **self.args)

                p = P(*self.index, _=[self._[-1] % i for i in other._], **self.args)
                p.name = f'{self} % {other}'
                return p
            if other.case in [PCase.NUM, PCase.NEGNUM]:
                # if other is a number, find the modulus of every value in self with the number
                p = P(*self.index, _=[i % other._[-1] for i in self._], **self.args)
                p.name = f'{self} % {other}'
                return p
            if len(self) != len(other):
                raise ValueError(
                    f"Index mismatch {self} % {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            # if both are just general parameters
            # zip through the values and find the modulus
            p = P(*self.index, _=[i % j for i, j in zip(self._, other._)], **self.args)
            p.name = f'{self} % {other}'
            return p

        # else let the other handle the modulus
        return self % other

    def __pow__(
        self,
        other: (
            V
            | Self
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> P:

        if other is None:
            # if raising to None, return self
            return self

        if self.case == PCase.ZERO:
            # if self is zero, return zero
            return 0

        if isinstance(other, (int, float)):
            # if raising to power of a number
            if other in [0, 0.0]:
                # if raising to zero, return 1
                return 1

            if other in [1, 1.0]:
                # if raising to one, return self
                return self
            if self.case in [PCase.NEGNUM, PCase.NUM]:
                # if self is a number, just find the power
                # let P handle the rest
                return P(*self.index, _=self._[-1] ** other, **self.args)

            # else find the power of each element in set raised to other
            p = P(*self.index, _=[i**other for i in self._], **self.args)
            p.name = f'{self}^({other})'
            return p

        if isinstance(other, list):
            if isinstance(other[0], tuple):
                raise ValueError(
                    "Cannot raise to a list of tuples. Please provide a valid parameter, variable, or number."
                )

            if self.case in [PCase.NUM, PCase.NEGNUM]:
                p = P(*self.index, _=[self._[-1] ** i for i in other], **self.args)
            else:
                if len(self) != len(other):
                    warn(
                        f"Index mismatch {self} ^ {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                    )
                # if list, just zip through the values and raise to power
                p = P(*self.index, _=[i**j for i, j in zip(self._, other)], **self.args)
            # change the name to add that something from a list has been added
            p.name = f'{self}^φ'
            return p

        if isinstance(other, tuple):
            # setting to the power of a parametric variable
            raise ValueError(
                f"Cannot raise {self} to a tuple. Please provide a valid parameter, variable, or number."
            )

        if isinstance(other, P):
            # other is a parameter set
            if other.case == PCase.ZERO:
                # if other is 0
                return 1

            if self.case in [PCase.NEGNUM, PCase.NUM]:
                # if self is a number, just find the power
                # let P handle the rest
                if other.case in [PCase.NUM, PCase.NEGNUM]:
                    return P(*self.index, _=self._[-1] ** other._[-1], **self.args)

                p = P(*self.index, _=[self._[-1] ** i for i in other._], **self.args)
                p.name = f'{self}^{other}'
                return p

            if other.case in [PCase.NUM, PCase.NEGNUM]:
                # if other is a number, raise every value in self to the power of the number
                p = P(*self.index, _=[i ** other._[-1] for i in self._], **self.args)
                p.name = f'{self}^{other}'
                return p

            if len(self) != len(other):
                raise ValueError(
                    f"Index mismatch {self} ^ {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            # if both are just general parameters
            # zip through the values and raise to power
            p = P(*self.index, _=[i**j for i, j in zip(self._, other._)], **self.args)
            p.name = f'{self}^{other}'
            return p

        # else let the other handle the power
        # not sure if this will ever be called tbh
        return self**other

    # -----------------------------------------------------
    #                    Relational
    # -----------------------------------------------------

    def __eq__(
        self,
        other: (
            Self
            | P
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> bool:

        if isinstance(other, (int, float)):

            if other in [0, 0.0]:
                if self.case == PCase.ZERO:
                    # if self is zero, return True
                    return True

            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                # if self is a number, compare it with the last value
                return self._[-1] == other
            # if not numeric, False
            raise NotImplementedError(
                f'{self} == {other}: cannot compare a parameter set with a number'
            )

        if isinstance(other, list):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                raise NotImplementedError(
                    f'{self} == {other}: cannot compare a number with a list.'
                )
            # if self is a set, compare it with the list
            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} == {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            return all([i == j for i, j in zip(self._, other)])

        if isinstance(other, P):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:

                if other.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                    # if self is a number, compare it with the last value of other
                    return self._[-1] == other._[-1]

                raise NotImplementedError(
                    f'{self} == {other}: cannot compare a number with a parameter set.'
                )

            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} == {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )

            return all([i == j for i, j in zip(self._, other._)])
        # else let other handle this
        return self == other

    def __le__(
        self,
        other: (
            Self
            | P
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> bool:

        if isinstance(other, (int, float)):

            if other in [0, 0.0]:
                if self.case in [PCase.ZERO, PCase.NEGNUM, PCase.NEGSET]:
                    # if self is zero or negative number, return True
                    return True

                if self.case in [PCase.NUM, PCase.SET]:
                    # if self is a positive number or set, return False
                    return False

            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                # if self is a number, compare it with the last value
                return self._[-1] <= other
            # if not numeric, False
            raise NotImplementedError(
                f'{self} <= {other}: cannot compare a parameter set with a number'
            )

        if isinstance(other, list):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                raise NotImplementedError(
                    f'{self} <= {other}: cannot compare a number with a list.'
                )
            # if self is a set, compare it with the list
            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} <= {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            return all([i <= j for i, j in zip(self._, other)])

        if isinstance(other, P):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:

                if other.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                    # if self is a number, compare it with the last value of other
                    return self._[-1] <= other._[-1]

                raise NotImplementedError(
                    f'{self} <= {other}: cannot compare a number with a parameter set.'
                )

            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} <= {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )

            return all([i <= j for i, j in zip(self._, other._)])
        # else let other handle this
        return self <= other

    def __ge__(
        self,
        other: (
            Self
            | P
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> bool:
        if isinstance(other, (int, float)):

            if other in [0, 0.0]:
                if self.case in [PCase.NEGNUM, PCase.NEGSET]:
                    # if self negative number or negated set, return False
                    return False

                if self.case in [PCase.ZERO, PCase.NUM, PCase.SET]:
                    # if self is a positive number or set, return True
                    return True

            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                # if self is a number, compare it with the last value
                return self._[-1] >= other
            # if not numeric, False
            raise NotImplementedError(
                f'{self} >= {other}: cannot compare a parameter set with a number'
            )

        if isinstance(other, list):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                raise NotImplementedError(
                    f'{self} >= {other}: cannot compare a number with a list.'
                )
            # if self is a set, compare it with the list
            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} >= {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            return all([i >= j for i, j in zip(self._, other)])

        if isinstance(other, P):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:

                if other.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                    # if self is a number, compare it with the last value of other
                    return self._[-1] >= other._[-1]

                raise NotImplementedError(
                    f'{self} >= {other}: cannot compare a number with a parameter set.'
                )

            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} >= {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )

            return all([i >= j for i, j in zip(self._, other._)])
        # else let other handle this
        return self >= other

    def __lt__(
        self,
        other: (
            Self
            | P
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> bool:
        if isinstance(other, (int, float)):

            if other in [0, 0.0]:

                if self.case in [PCase.NEGNUM, PCase.NEGSET]:
                    # if self negative number or negated set, return True
                    return True

                if self.case in [PCase.ZERO, PCase.NUM, PCase.SET]:
                    # if self is a positive number or set, return True
                    return False

            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                # if self is a number, compare it with the last value
                return self._[-1] < other
            # if not numeric, False
            raise NotImplementedError(
                f'{self} < {other}: cannot compare a parameter set with a number'
            )

        if isinstance(other, list):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                raise NotImplementedError(
                    f'{self} < {other}: cannot compare a number with a list.'
                )
            # if self is a set, compare it with the list
            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} < {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            return all([i < j for i, j in zip(self._, other)])

        if isinstance(other, P):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:

                if other.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                    # if self is a number, compare it with the last value of other
                    return self._[-1] < other._[-1]

                raise NotImplementedError(
                    f'{self} < {other}: cannot compare a number with a parameter set.'
                )

            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} < {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )

            return all([i < j for i, j in zip(self._, other._)])
        # else let other handle this
        return self < other

    def __ne__(
        self,
        other: (
            Self
            | P
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> bool:
        if isinstance(other, (int, float)):

            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                # if self is a number, compare it with the last value
                return self._[-1] != other
            # if not numeric, False
            raise NotImplementedError(
                f'{self} != {other}: cannot compare a parameter set with a number'
            )

        if isinstance(other, list):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                raise NotImplementedError(
                    f'{self} != {other}: cannot compare a number with a list.'
                )
            # if self is a set, compare it with the list
            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} != {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            return all([i != j for i, j in zip(self._, other)])

        if isinstance(other, P):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:

                if other.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                    # if self is a number, compare it with the last value of other
                    return self._[-1] != other._[-1]

                raise NotImplementedError(
                    f'{self} != {other}: cannot compare a number with a parameter set.'
                )

            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} != {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )

            return all([i != j for i, j in zip(self._, other._)])
        # else let other handle this
        return self != other

    def __gt__(
        self,
        other: (
            Self
            | P
            | T
            | F
            | int
            | float
            | tuple[int | float]
            | list[int | float | tuple[int | float]]
            | None
        ),
    ) -> bool:
        if isinstance(other, (int, float)):

            if other in [0, 0.0]:

                if self.case in [PCase.ZERO, PCase.NEGNUM, PCase.NEGSET]:
                    # if self negative number or negated set, return True
                    return False

                if self.case in [PCase.NUM, PCase.SET]:
                    # if self is a positive number or set, return True
                    return True

            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                # if self is a number, compare it with the last value
                return self._[-1] > other
            # if not numeric, False
            raise NotImplementedError(
                f'{self} > {other}: cannot compare a parameter set with a number'
            )

        if isinstance(other, list):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                raise NotImplementedError(
                    f'{self} > {other}: cannot compare a number with a list.'
                )
            # if self is a set, compare it with the list
            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} > {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )
            return all([i > j for i, j in zip(self._, other)])

        if isinstance(other, P):
            if self.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:

                if other.case in [PCase.NUM, PCase.NEGNUM, PCase.ZERO]:
                    # if self is a number, compare it with the last value of other
                    return self._[-1] > other._[-1]

                raise NotImplementedError(
                    f'{self} > {other}: cannot compare a number with a parameter set.'
                )

            if len(self) != len(other):
                warn(
                    f"Index mismatch {self} > {other}: len(self) ({len(self)}) != len(other) ({len(other)})"
                )

            return all([i > j for i, j in zip(self._, other._)])
        # else let other handle this
        return self > other

    # -----------------------------------------------------
    #                    Vector
    # -----------------------------------------------------

    def __iter__(self) -> Self:
        return iter(self._)

    def __len__(self):
        return len(self.map)

    def __call__(self, *key: I) -> Self:

        def lister(inp: tuple[I]) -> tuple[I | list[V]]:
            return tuple([i] if isinstance(i, V) else i for i in inp)

        # if a dependent variable is being passed in the key
        # extract variable from the index (it will be in a list)
        def delister(inp: tuple[I | list[V]]):
            return tuple(i[0] if isinstance(i, list) else i for i in inp)

        if not key or delister(key) == delister(self.index):
            # if the index is an exact match
            # or no key is passed
            return self

        # the check helps to handle if a variable itself is an index
        # we do not want to iterate over the entire variable set
        # but treat the variable as a single index element
        key: tuple[I] | set[tuple[I]] = lister(key)

        # if a subset is passed,
        # first create a product to match
        # the indices

        # create a new variable set to return
        p = P(**self.args)
        p.name, p.n = self.name, self.n
        p.index = key

        # should be able to map these
        for index in product(*key):
            # this helps weed out any None indices
            # i.e. skips
            if any(i is None for i in index):
                index = None

            if index is None:
                parameter = None
            else:
                parameter = self.map[index]

            p.map[index] = parameter
            p._.append(parameter)

        return p

    def __getitem__(self, pos: int) -> float | int:
        return self._[pos]

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    def __str__(self):
        return rf'{self.name}'

    def __repr__(self):
        return str(self.name)

    def __hash__(self):
        return hash(str(self.name))

    # -----------------------------------------------------
    #                    Export
    # -----------------------------------------------------

    def sympy(self):
        """symbolic representation"""
        if has_sympy:
            return IndexedBase(str(self))[
                symbols(",".join([f'{d}' for d in self.index]), cls=Idx)
            ]
        print(
            "sympy is an optional dependency, pip install gana[all] to get optional dependencies"
        )

    def pyomo(self):
        """Pyomo representation"""
        # idx = [i.pyomo() for i in self.index]
        # return PyoParam(*idx, initialize=self._, doc=str(self))
        if has_pyomo:
            return PyoParam(
                initialize=self._,
                doc=str(self),
            )
        print(
            "pyomo is an optional dependency, pip install gana[all] to get optional dependencies"
        )



"""Function Compositions"""

from ..sets.function import F
from ..sets.objective import O


def inf(function: F) -> F:
    """Minimize the function"""
    return O(function=function)


def sup(function: F) -> F:
    """Maximize the function"""
    return O(function=-function)



"""Zero"""

from typing import Self


class Z:
    """Zero, Shunya"""

    def __init__(self, _: float = 0, neg: bool = False):
        if _ and _ < 0:
            raise ValueError('Zero value cant be negative, give neg = True')
        # a tolerance value if needed
        self._ = _
        self.neg = neg

    @property
    def name(self):
        """name"""
        if self.neg:
            return '-0'
        else:
            return '0'

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __len__(self):
        return 1

    def __neg__(self):
        if self.neg:
            return Z()
        else:
            return Z(neg=True)

    def __pos__(self):
        return self

    def __add__(self, other: Self | float):
        return other

    def __radd__(self, other: Self | float):
        return other

    def __sub__(self, other: Self | float):
        if isinstance(other, Z):
            return self
        else:
            return -other

    def __rsub__(self, other: Self | float):
        return other

    def __mul__(self, other: Self | float):
        return self

    def __rmul__(self, other: Self | float):
        return self

    def __truediv__(self, other: Self | float):
        return self

    def __rtruediv__(self, other: Self | float):
        raise ValueError('Cant divide by zero')

    def __gt__(self, other: Self | float):

        if isinstance(other, Z):
            if self.neg and not other.neg:
                return True

            if not self.neg and other.neg:
                return False

            if self.neg and other.neg:
                return False

            if not self.neg and not other.neg:
                return False

        if isinstance(other, (int, float)):

            if other < 0:
                return True

            if other >= 0:
                return False

    def __ge__(self, other: Self | float):
        return self > other

    def __lt__(self, other: Self | float):
        return not self > other

    def __le__(self, other: Self | float):
        return not self > other

    def __eq__(self, other: Self | float):
        if isinstance(other, Z):
            if self.neg and other.neg:
                return True

            if not self.neg and not other.neg:
                return True

            if (not self.neg and other.neg) or (self.neg and not other.neg):
                return False
        else:
            return False

    def __ne__(self, other: Self | float):
        return not self == other

    # def __call__(self) -> IndexedBase:
    #     """Symbol"""
    #     return -IndexedBase('δ') if self.neg else IndexedBase('δ')



"""Objective Function"""

from __future__ import annotations

from typing import TYPE_CHECKING

from IPython.display import Math, display

if TYPE_CHECKING:
    from .function import F


class O:
    """Objective Function"""

    def __init__(self, function: F):

        if all([v.parent is None for v in function.variables]):
            # if the function is defined using a variable element
            self.function = function[0]
            # the A matrix becomes the C matrix

            self.C = self.function.A

        else:
            # if the function is defined using variable sets
            self.function = function
            # the A matrix becomes the C matrix
            self.C = self.function.A[0]

        self.C = self.C[0] if isinstance(self.C[0], list) else self.C
        self._ = self.function

        # The index is taken from the function
        self.index = self.function.index
        # so are the variables

        self.variables = self.function.variables
        # the name, always minimization in gana
        self.name = rf'min({self.function})'
        # whats contained in the set

        # value obtained after optimization
        self.X: float = None

        # number in the program
        self.n = None

        # name given by user in program
        self.pname: str = None

    def update_variables(self):
        """Informs that the variables are optimized by this objective."""
        for v in self.variables:
            v.min_by.append(self)

    @property
    def P(self):
        """Variable positions"""
        return [v.n for v in self.variables]

    @property
    def matrix(self) -> dict:
        """Matrix as dict"""
        return self.function.matrix

    def sol(self, asfloat: bool = False):
        """Solution"""
        if asfloat:
            return self.X

        display(Math(self.latex() + r'=' + rf'{self.X}'))

    def latex(self):
        """Latex representation"""
        return rf'min \hspace{{0.2cm}} {self.function.latex()}'

    def show(self):
        """Pretty Print"""
        display(Math(self.latex()))

    def mps(self):
        """Name in MPS file"""
        return f'O{self.n}'

    def __str__(self):
        return rf'{self.name}'

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))
